#include "MyConfigImpl.h"
#include <qfont.h>
#include <qfontdialog.h>
#include <qslider.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qspinbox.h>
#include <registry.h>
#include <wine/winreg.h>
#include <stdio.h>

/* 
 *  Constructs a MyConfigDialog_impl which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
MyConfigDialog_impl::MyConfigDialog_impl( QWidget* parent,  const char* name, bool modal, WFlags fl )
    : MyConfigDialog( parent, name, modal, fl )
{
}
MyConfigDialog_impl::MyConfigDialog_impl() : MyConfigDialog(0, "Configuration", true)
{
    float value=ReadAsync();
    AsyncSlider->setValue(int(value*100.));
    onAsyncValChanged(0);

    m_defFont=ReadFont();
    char str[250];
    sprintf(str, "%s, %d pt%s%s",
    m_defFont.family().ascii(), m_defFont.pointSize(),
    m_defFont.bold()?", bold":"",
    m_defFont.italic()?", italic":"");

    FontType->setText(str);
    Preview->setFont(m_defFont);
    
    float sub_async=ReadSubAsync();
    if(sub_async<0)
    {
	CheckBox1->setChecked(true);
	sub_async*=-1;
    }
    else CheckBox1->setChecked(false);
    m_pSubAsyncMin->setValue(int(sub_async/60));
    sub_async-=60*int(sub_async/60);
    m_pSubAsyncSec->setValue(int(sub_async));
    _hwaccel->setChecked(ReadYUV());
        
    connect(Ok, SIGNAL(clicked()), this, SLOT(onOk()));
    connect(Cancel, SIGNAL(clicked()), this, SLOT(onCancel()));
    connect(m_pFontButton, SIGNAL(clicked()), this, SLOT(onFont()));
    connect(AsyncSlider, SIGNAL(valueChanged(int)), this, SLOT(onAsyncValChanged(int)));
} 

void MyConfigDialog_impl::onOk()
{
    float value=(float)AsyncSlider->value();
    WriteAsync(value/100.f);
    WriteFont(m_defFont);
    float sub_async=m_pSubAsyncMin->value()*60+m_pSubAsyncSec->value();
    if(CheckBox1->isChecked())sub_async*=-1;
    WriteSubAsync(sub_async);
    WriteYUV(_hwaccel->isChecked());
    accept();
}


void MyConfigDialog_impl::onAsyncValChanged(int)
{
    float value=(float)AsyncSlider->value();
    value/=100.;
    char text[100];
    sprintf(text, "%1.2f s", value);
    Async->setText(text);
}

void MyConfigDialog_impl::onFont()
{
    bool isOk;
    QFont font=QFontDialog::getFont(&isOk, m_defFont, this, "Select font");
    if(isOk)
    {
        m_defFont=font;
        Preview->setFont(font);

        char str[250];
        sprintf(str, "%s, %d pt%s%s",
	    m_defFont.family().ascii(),
	    m_defFont.pointSize(),
	    m_defFont.bold()?", bold":"",
	    m_defFont.italic()?", italic":"");
	    FontType->setText(str);

        update();
    }    
}


/*  
 *  Destroys the object and frees any allocated resources
 */
MyConfigDialog_impl::~MyConfigDialog_impl()
{
    // no need to delete child widgets, Qt does it all for us
}

#include "MyConfig.moc"
#include "MyConfigImpl.moc"
