#include "mywidget.h"
#include <qdialog.h>
#include <qslider.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <videodecoder.h>
using namespace std;

extern DrawEvent g_event;

class QConfDialog: public QDialog
{
    Q_OBJECT
    
    const CodecInfo& info;
    const vector<AttributeInfo>& attrs;
    vector<QObject*> items;
    IRtConfig* rt;
    QPushButton* b_apply;
public:
    QConfDialog(const CodecInfo& _info, IRtConfig* _rt=0);
public slots:    
    void accept();
    void apply();
    void valueChanged(int);
};

QConfDialog::QConfDialog(const CodecInfo& _info, IRtConfig* _rt=0)
    :info(_info), rt(_rt), QDialog(0, "", true), attrs(_info.decoder_info),
	b_apply(0)
{
    setCaption("Video decoder configuration");
    int y=10;
    int val;
    
    for(int i=0; i<attrs.size(); i++)
    {
	QSlider* slider;
	QGroupBox* box;
	QLabel* label;
	char txt[20];
	if(attrs[i].kind==AttributeInfo::Integer)
	{
	    if(attrs[i].i_min+1>=attrs[i].i_max)
		continue;//integers with unlimited values 
			 //and boolean attributes are not supported
	    box=new QGroupBox(this, "");
	    box->setTitle(attrs[i].about.c_str());
	    box->move(10, y);
	    box->resize(240, 75);
	    
	    slider=new QSlider(box, "Slider");
	    items.push_back(static_cast<QObject*>(slider));
	    slider->move(20,15);
	    slider->resize(200,30);
	    slider->setOrientation(QSlider::Horizontal);
	    slider->setTickmarks(QSlider::Both);
	    slider->setMinValue(attrs[i].i_min);
	    slider->setMaxValue(attrs[i].i_max);
	    if(attrs[i].i_max-attrs[i].i_min<20)
	    {
		slider->setLineStep(1);
	        slider->setPageStep(1);
	    }
	    else
	    {
		slider->setLineStep((attrs[i].i_max-attrs[i].i_min)/10);
		slider->setPageStep((attrs[i].i_max-attrs[i].i_min)/10);
	    }
	    if(IVideoDecoder::GetExtendedAttr(info, attrs[i].name.c_str(), val)==0)
		slider->setValue(val);
	    if(rt)
		connect(slider, SIGNAL(valueChanged(int)), this, SLOT(valueChanged(int)));
	    sprintf(txt, "%d", attrs[i].i_min);
	    label=new QLabel(box, "");
	    label->setText(txt);
	    label->move(20, 45);
	    label->resize(20, 20);
	    sprintf(txt, "%d", attrs[i].i_max);
	    label=new QLabel(box, "");
	    label->setText(txt);
	    label->move(210, 45);
	    label->resize(20, 20);
	    y+=85;
	}
	//other types of atributes are not supported
    }
    resize(260,y+40);
    QPushButton* b_ok=new QPushButton("Ok", this);
    b_ok->resize(75,30);
    b_ok->move(10, y);
    connect(b_ok, SIGNAL(clicked()), this, SLOT(accept()));

    QPushButton* b_cancel=new QPushButton("Cancel", this);
    b_cancel->resize(75,30);
    b_cancel->move(175, y);
    connect(b_cancel, SIGNAL(clicked()), this, SLOT(reject()));
    
    if(rt)
    {
	b_apply=new QPushButton("Apply", this);
	b_apply->resize(80, 30);
	b_apply->move(90, y);
	connect(b_apply, SIGNAL(clicked()), this, SLOT(apply()));
	b_apply->setEnabled(0);
    }
}
void QConfDialog::valueChanged(int)
{
    if(b_apply)
	if(!b_apply->isEnabled())
	    b_apply->setEnabled(1);
}


void QConfDialog::accept()
{
    apply();
    return QDialog::accept();
}

void QConfDialog::apply()
{
    for(int i=0; i<attrs.size(); i++)
    {
        QSlider* slider;
        if(slider=dynamic_cast<QSlider*>(items[i]))
	{
	    IVideoDecoder::SetExtendedAttr(info, attrs[i].name.c_str(), slider->value());
	    if(rt)rt->SetValue(attrs[i].name.c_str(), slider->value());
	}
    }
}

void MyWidget::decoder_config()
{
    if(!player->isValid())
	return;
    IRtConfig* rt=player->GetRuntimeConfig();
    QConfDialog dlg(player->GetCodecInfo(), rt);
    if(dlg.exec()==QDialog::Accepted)
    {
	_in_reseek=1;
	if(!rt)
    	    player->Restart();//this call blocks until main processing thread
	//is stopped, so it can safely restart decoder.
	//Operations with _in_reseek and mutex guarantee that
	//we won't receive a deadlock.
	_in_reseek=0;
    }
}
#include "decoder_config.moc"