/********************************************************

	Main AVI player widget
	Copyright 2000 Eugene Kuznetsov (divx@euro.ru)

*********************************************************/


#define QT_THREAD_SUPPORT


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include <qapp.h>
#include <qfiledialog.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>

#include "renderer.h"
#include "mywidget.h"
#include "MyConfigImpl.h"

#include <default.h>
//#include <X11/Xlib.h>

extern "C" 
{
    int XQueryPointer(
    Display*            /* display */,
    HANDLE              /* w */,
    HANDLE*             /* root_return */,
    HANDLE*             /* child_return */,
    int*                /* root_x_return */,
    int*                /* root_y_return */,
    int*                /* win_x_return */,
    int*                /* win_y_return */,
    unsigned int*       /* mask_return */
    );
};
				    
using namespace std;

extern MyWidget* m;

#define Q_MenuEvent QEvent::User
#define Q_PosEvent QEvent::Type(QEvent::User+1)
#define Q_DrawEvent QEvent::Type(QEvent::User+2)
#define Q_SubEvent QEvent::Type(QEvent::User+3)
#define Q_MyCloseEvent QEvent::Type(QEvent::User+4)
void MyWidget::menu_slot()
{
    QEvent* e=new QEvent(Q_MenuEvent);
    qApp->postEvent(m, e);
    qApp->wakeUpGuiThread();
}
void MyWidget::pos_slot()
{
    QEvent* e=new QEvent(Q_PosEvent);
    qApp->postEvent(m, e);
    qApp->wakeUpGuiThread();
}
void MyWidget::draw_func()
{
    QEvent* e=new QEvent(Q_DrawEvent);
    qApp->postEvent(m, e);
    qApp->wakeUpGuiThread();
}
void MyWidget::drawSubtitles_slot()
{
    QEvent* e=new QEvent(Q_SubEvent);
    qApp->postEvent(m, e);
    qApp->wakeUpGuiThread();    
}
void MyWidget::quit_func()
{
    QEvent* e=new QEvent(Q_MyCloseEvent);
    qApp->postEvent(m, e);
    qApp->wakeUpGuiThread();
}
bool MyWidget::event(QEvent* e)
{
    switch(e->type())
    {
    case Q_MenuEvent:
	menu_func();
	return true;
    case Q_PosEvent:
	updatePos();
	return true;
    case Q_DrawEvent:
	draw_slot();
	return true;
    case Q_SubEvent:
	drawSubtitles();
	return true;
    case Q_MyCloseEvent:
	close(true);
	return true;
    }	
    return QWidget::event(e);
}
void MyWidget::menu_func()
{
    HANDLE root_return, child_return;
    int root_x_return, root_y_return;
    int win_x_return, win_y_return;
    unsigned int mask_return;

    XQueryPointer(x11Display(), handle(), &root_return, &child_return, &root_x_return, &root_y_return,
    &win_x_return, &win_y_return, &mask_return);
	
    QPopupMenu* popup=new QPopupMenu;
    popup->insertItem("Zoom 2x", this, SLOT(zoom_2x()));
    popup->insertItem("Zoom 1x", this, SLOT(zoom_1x()));
    popup->insertItem("Zoom 0.5x", this, SLOT(zoom_0_5x()));
#ifdef USE_SDL
    popup->insertSeparator();
    popup->insertItem("Fullscreen", this, SLOT(fullscreen()));
#ifdef HAVE_LIBXXF86VM
    popup->insertItem("Maximize", this, SLOT(maximize()));
#endif
#endif    
    popup->insertSeparator();
    popup->insertItem("Properties", this, SLOT(about_movie()));
    
    if(player->GetCodecInfo().decoder_info.size())
    {
	popup->insertSeparator();
	popup->insertItem("Decoder config", this, SLOT(decoder_config()));
    }
    popup->popup(QPoint(root_x_return, root_y_return));
}
void MyWidget::zoom_2x()
{
    if(!player->isValid())return;
    
    int w=2*player->width();
    int h=2*player->height();
    
//    renderer->Resize(w,h);
//    
#ifndef USE_SDL
    if(player->HasSubtitles())
        h+=100;
    else
        h+=50;
	
    resize(w, h);
#else
    renderer->Resize(w,h);
#endif        
}
void MyWidget::zoom_1x()
{
    if(!player->isValid())return;
    
    int w=player->width();
    int h=player->height();
    
#ifndef USE_SDL
    if(player->HasSubtitles())
        h+=100;
    else
        h+=50;
	
    resize(w, h);
#else
    renderer->Resize(w,h);
#endif        
}
void MyWidget::zoom_0_5x()
{
    if(!player->isValid())return;
    if(!renderer)return;
    
    int w=player->width()/2;
    int h=player->height()/2;
    
#ifndef USE_SDL
    if(player->HasSubtitles())
        h+=100;
    else
        h+=50;
	
    resize(w, h);
#else
    renderer->Resize(w,h);
#endif        
}
void MyWidget::fullscreen()
{
    if(!player->isValid())return;
    if(!renderer)return;
    renderer->toggle_fullscreen();
}
void MyWidget::maximize()
{
    if(!player->isValid())return;
    if(!renderer)return;
    renderer->toggle_fullscreen(true);
}
void MyWidget::showconf_func()
{
    MyConfigDialog_impl dlg;
    dlg.exec();
    readConfig();
}

void MyWidget::paintEvent(QPaintEvent *)
{
    if(!player)
    {
	drawBackground();
	return;
    }
    if((!player->isValid()) || (player->isStopped()))
    {
    	drawBackground();
	return;
    }	
    updatePos();
    return;
}

void MyWidget::rcreseek(double pos)
{
    scroller->setValue(pos);
    player->reseek_exact(pos);
    updatePos();
}

void MyWidget::stopfunc(int)
{
    m->scroller->setValue(0);
    m->drawBackground();
}

void MyWidget::open_func()
{
    QString open_qs;
    string path=Registry::ReadString("aviplay", "file", ".");
    open_qs=QFileDialog::getOpenFileName(path.c_str(), "AVI files (*.avi *.AVI);;ASF files (*.asf *.ASF);;All files(*)");
//    open_qs=QFileDialog::getOpenFileName(QString::null,
//	 "AVI files (*.avi *.AVI);;ASF files (*.asf *.ASF);;All files (*.*)");
    if(open_qs.isNull())
	return;
    try
    {
	initPlayer(open_qs);
	Registry::WriteString("aviplay", "file", open_qs.ascii());
    }	
    catch(FatalError& error)
    {
	QMessageBox::information(this, 
	    QString("Error in ")+QString(error.GetModule())+QString(" module"),
	    error.GetDesc(), QMessageBox::Ok);
	error.PrintAll();
	int w=384;	
	int h=0;
#ifndef USE_SDL	
	resize(w,h+30);
#else
	setFixedSize(w,h+30);
#endif		
	moveItems(w,h+50);

	repaint();
    }	
}
MyWidget::~MyWidget()
{
    endPlayer();
    if(player)delete player;
}

#define __MODULE__ "AVI player widget"

void MyWidget::keyPressEvent(QKeyEvent * event)
{
    // well-known x11amp key events: 'x' stop, 'c' pause, 'v' play.
    // for SDL events are also mirrored in EventFilter ( see renderer.cpp )
    switch(event->ascii())
    {
	case 'Q':
	case 'q':
	    quit_func();
	    break;
	case 'x':
	case 'X':
	    stop_func();
	    break;
	case 'c':
	case 'C':
	    pause_func();
	    break;
	case 'v':
	case 'V':
	    play_func();
	    break;
    }	            
}
void MyWidget::mouseDoubleClickEvent(QMouseEvent * event)
{
    showconf_func();
}

void MyWidget::resizeEvent(QResizeEvent* ev)
{
#ifndef USE_SDL
    printf("Resize()\n");
    const QSize& s=ev->size();
    int picw=s.width();
    int pich=s.height();
    if(player==0)
    {
	printf("player=0\n");
	goto fail;
    }	
    if(renderer==0)
    {
	printf("renderer=0\n");
	goto fail;
    }	
    if(player->isValid()==0)goto fail;
    moveItems(picw, pich);
    if(player->HasSubtitles())
	pich-=100;
    else
	pich-=50;
    if(pich<16)
	return;
    if(picw<16)
	return;
    if(pich<player->height()/4)
	return;
    if(picw<player->width()/4)
	return;
    renderer->Resize(picw, pich);
fail:
;
    QWidget::resizeEvent(ev);
#endif    
}	
void MyWidget::mouseReleaseEvent( QMouseEvent * e)
{
#ifndef USE_SDL
    if(e->button()==RightButton)
	menu_func();
#endif
    return QWidget::mouseReleaseEvent(e);
}

void MyWidget::about_func()
{
    QPopupMenu* popup=new QPopupMenu;
    popup->insertItem("Usage", this, SLOT(help()));
    popup->insertItem("About", this, SLOT(about()));
    popup->insertItem("Config", this, SLOT(showconf_func()));
#ifndef USE_SDL    
    int w=width();
    int h=height();
#else
    int w=384;
    int h=50;
#endif        
    HANDLE root_return, child_return;
    int root_x_return, root_y_return;
    int win_x_return, win_y_return;
        unsigned int mask_return;

    XQueryPointer(x11Display(), handle(), &root_return, &child_return, &root_x_return, &root_y_return,
    &win_x_return, &win_y_return, &mask_return);
	
    popup->popup(QPoint(root_x_return, root_y_return));
}
void MyWidget::Resize(int w,int h)
{
    if(renderer)renderer->Resize(w,h);
}
void MyWidget::remote_command()
{
    char command[32];
    int command_pos = 0;
    fd_set fds; // stdin changed?
    struct timeval tv; // how long to w8
    int n; // indicating whether something happened on stdin

     printf ("Pos: %i s / %i s\n", (long) m->GetPos(),
	(long) m->GetVideoLength());
     tv.tv_sec = 0;
     tv.tv_usec = 50000;
     FD_ZERO(&fds);
     FD_SET(STDIN_FILENO, &fds);
     if ((n = select(32, &fds, NULL, NULL, &tv)) != 0)
     {
       double pos;
  int i;
  int complete = 0;
 
  while (command_pos < 32 &&
 (i = read (STDIN_FILENO, command + command_pos, 1)) > 0 &&
 command[command_pos] != '\r' &&
 command[command_pos] != '\n') {
 command_pos++;
  }
  if (command_pos == 31 || command[command_pos] == '\r' ||
 command[command_pos] == '\n') {
 command[command_pos] = 0;
 complete = 1;
  }
      if (complete == 1) {
     if (command[0] == 'p' || command[0] == 'P')
       pause_func();
     else if (command[0] == 's' || command[0] == 'S')
       stop_func();
     else if (command[0] == 'q' || command[0] == 'Q')
       quit_func();
     else if (command[0] == 'r' || command[0] == 'R') {
       for (i = 1; i < 32 && command[i] >= '0' && command[i] <= '9'; ++i)
           ;
     command[i] = 0;
     pos = atof (command + 1);
     rcreseek(pos);
     } else {
       printf ("Unknown command: %s\n", command);
     }
     command_pos = 0;
     complete = 0;
    }
    return;
    }
}

void MyWidget::about_movie()
{
    char t[128];
    if(!player->isValid())return;
    string s="File ";
    s+=player->fileName();
    s+="<br><br>";
    if(player->GetAudioLength()>0)
    {
	sprintf(t, ", %.2f seconds.<br>", player->GetAudioLength());
	s+="Audio stream: ";
	s+=player->GetAudioFormat();
	s+=t;
    }
    if(player->GetVideoLength()>0)
    {
	sprintf(t, "Video stream: %dx%d, %d fps, ", player->width(), player->height(), player->fps());
	s+=t;
	s+=player->GetVideoFormat();
	sprintf(t, ", %.2f seconds.<br>", player->GetVideoLength());
	s+=t;
    }
    QMessageBox::information(this, "About file", s.c_str());
}
		     
			     
#include "mywidget.moc"
