#include <qwidget.h>
#include <qslider.h>
#include <qpushbutton.h>
//#include "qvolumebar.h"

#include <aviplay.h>
#include <pthread.h>
#include "playerwidget.h"
class VideoRenderer;
class QPixmap;
class QTimer;
class MyConfigDialog;
class QLabel;
class PixmapButton: public QPushButton
{
    QPixmap m_pm;
public:    
    PixmapButton(QPixmap& pm, const char* name, QWidget* parent):QPushButton(name,parent), m_pm(pm){}
protected:    
    void paintEvent(QPaintEvent *qp);
};
class MyWidget : public QWidget, public PlayerWidget
{
	Q_OBJECT
public:	
	static void drawfunc(CImage*);
	static void stopfunc(int);
protected:
	void paintEvent(QPaintEvent *);
	void keyPressEvent(QKeyEvent *);
	void mouseDoubleClickEvent(QMouseEvent*);
	void mouseReleaseEvent( QMouseEvent * );
    	void readConfig();
	void resizeEvent(QResizeEvent *);
	IAviPlayer* player;
	pthread_mutex_t mutex;
public:
    virtual void PW_showconf_func() {return showconf_func();}
    virtual void PW_stop_func() {return stop_func();}
    virtual void PW_pause_func() {return pause_func();}
    virtual void PW_play_func() {return play_func();}
    virtual void PW_quit_func() {return quit_func();}
    virtual void PW_menu_slot() {return menu_slot();}
    virtual void PW_fullscreen() {return fullscreen();}
    virtual void PW_Resize(int w, int h) {return Resize(w,h);}
	
	
public:
	MyWidget(int, int);
	~MyWidget();
	int initPlayer(const char* filename, const char* subname=NULL, int=0, int=0);
	void endPlayer();
	void updatePos();
	void updatePosDisplay(int pos);
	void drawSubtitles();
	int isValid(){return player->isValid();}
	int isStopped(){return player->isStopped();}
	int isPlaying(){return player->isPlaying();}
	int isPaused(){return player->isPaused();}
	double GetPos(){return player->GetPos();}
	double GetVideoLength(){return player->GetVideoLength();}
	void drawBackground();
	int is_tracking;
	int _in_reseek;
public slots:
	void nextpage_func()
	{
	    if(player->isValid())
	    {
		if(player->page_up()==0)
		    updatePos();
	    }	
        }
	void prevpage_func()
	{
	    if(player->isValid())
		if(player->page_down()==0)
		    updatePos();
	}
	void slider_func(){is_tracking=1;}
	void reseek_func();
	void moved_func(int);
	void rcreseek(double pos);
	void pause_func()
	{
	    if(player->isValid())
	    {
		if(player->isPaused())
		    player->pause(false);
    		else
		    player->pause(true);
	    }	    
	}
	void open_func();
	void stop_func()
	{
	    if(player->isValid())
	    {
		player->stop();
		scroller->setValue(0);
		drawBackground();
	    }	
	}    
	void play_func()
	{
	    if(player->isValid())
	    {
		if(!player->isPlaying())
		    player->start();
		else
		    player->pause(false);
	    }
	}    	
	void showconf_func();
	void menu_func();
	void zoom_0_5x();
	void zoom_1x();
	void zoom_2x();
	void fullscreen();
	void maximize();
	void moveItems(int,int);
	void about_func();
	void quit_func();
	void Resize(int,int);

	void menu_slot();
	void pos_slot();
	void remote_command();
	
	bool event(QEvent* e);
	
	void help();
	void about();
	void about_movie();
	void decoder_config();
	void dropEvent( QDropEvent * );
	void dragEnterEvent(QDragEnterEvent* event);
	void draw_func();
	void draw_slot();
	void drawSubtitles_slot();
protected:
	QSlider* scroller;
//	QVolumeBar* vbar;
	PixmapButton *b_play;
	PixmapButton *b_open;
	PixmapButton *b_pause;
	PixmapButton *b_stop;
	PixmapButton *b_about;
	QLabel* l_drop;
	QLabel* l_time;
	MyConfigDialog* dlg;
	
	VideoRenderer* renderer;
};
struct DrawEvent
{
    CImage* data;
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    DrawEvent():data(0)
    {
	pthread_mutex_init(&mutex, 0);
	pthread_cond_init(&cond, 0);
    }	
    ~DrawEvent()
    {
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&cond);
    }
};
