/********************************************************

	Video renderers
	Copyright 2000 Eugene Kuznetsov (divx@euro.ru)

*********************************************************/
#include <config.h>


#include <unistd.h>
#define QT_CLEAN_NAMESPACE
#ifdef __FreeBSD__
#include <machine/param.h>
#include <sys/types.h>
#endif
#include <sys/ipc.h>
#include <sys/shm.h>                               

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <pthread.h>
#include <errno.h>
#include "renderer.h"
#include "RegAccess.h"

#include <qapp.h>
#include <qwidget.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#ifdef HAVE_LIBXXF86DGA
#include <X11/extensions/xf86dga.h>
#endif

#ifdef HAVE_LIBXXF86VM
#include <X11/extensions/xf86vmode.h>
#endif

#ifdef USE_SDL
/*
#ifdef USE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#endif
*/
#include <SDL.h>
#include <SDL_video.h>
#include <SDL_syswm.h>
#include <SDL_mouse.h>
#endif

#include <except.h>
#include "playerwidget.h"
#define __MODULE__ "VideoRenderer"
int width, height;
//code taken from kwintv
int GetPhysicalDepth(QPaintDevice* dev)
{
        Display *dpy = dev->x11Display();
        Visual *vis = (Visual*)dev->x11Visual();
	int n, pixmap_bits;
	XPixmapFormatValues *pf=XListPixmapFormats(dpy, &n);
	for ( int i = 0; i<n; i++)
	{
		if(pf[i].depth==QColor::numBitPlanes())
		pixmap_bits=pf[i].bits_per_pixel;
	}
	if(pixmap_bits==16)
	{
	    if(vis->red_mask==0x7c00)
		pixmap_bits=15;
	}	
	return pixmap_bits;
}

inline static void qSafeXDestroyImage( XImage *x )
{
    if ( x->data ) {
	free( x->data );
	x->data = 0;
    }
    XDestroyImage( x );
}



class ShmRenderer : public VideoRenderer
{
    bool	       xshminit;
    XShmSegmentInfo xshminfo;
    XImage	      *xshmimg;
    Pixmap	       xshmpm;
    GC 			gc;
    pthread_mutex_t mutex;
    int xpos;
public:
    ShmRenderer() : xshminit(0), xshmimg(0), xshmpm(0)
    {
    	pthread_mutex_init(&mutex, 0);
    }
    void Alloc()
    {
        static int major, minor;
	static Bool pixmaps_ok;
        Display *dpy = dev->x11Display();
        int dd	 = dev->x11Depth();
        Visual *vis	 = (Visual*)dev->x11Visual();
	printf("Creating SHM renderer, width %d, height %d\n", pic_w, pic_h);
	printf("Server vendor %s, release %d\n", ServerVendor(dpy), VendorRelease(dpy));
	
	try
	{
    	    XGCValues xcg;
	    xcg.graphics_exposures=false;
	    gc=XCreateGC(dpy, dev->handle(), GCGraphicsExposures, &xcg);
	    if ( !XShmQueryVersion(dpy, &major, &minor, &pixmaps_ok) )
		throw FATAL("MIT SHM extension not supported");

	    bool ok;
	    int _pic_w=((pic_w+m_w-1)/m_w)*m_w;
	    int _pic_h=((pic_h+m_h-1)/m_h)*m_h;
    	    xshminfo.shmid = shmget( IPC_PRIVATE,
    			     _pic_w*_pic_h*4,
    			     IPC_CREAT | 0777 );
	    ok = xshminfo.shmid != -1;
	    if(!ok)
		throw FATAL("Can't get shared memory segment");
    	
	    xshminfo.shmaddr = (char*)shmat( xshminfo.shmid, 0, 0 );
	    ok = xshminfo.shmaddr != 0;
	    xshminfo.readOnly = FALSE;

    	    if ( !ok )
		throw FATAL("Can't attach shared memory segment");
	    
	    ok = XShmAttach( dpy, &xshminfo );
	    if ( !ok )
		throw FATAL("XShmAttach failed");
	    
    	    xshmimg = XShmCreateImage( dpy, vis, dd, ZPixmap, xshminfo.shmaddr, &xshminfo, _pic_w, _pic_h );

	    if ( !xshmimg )
	        throw FATAL("Can't create shared image");

    	    printf("Video device: %d bpp ( physical %d ), color masks %X,%X,%X\n",
		dev->x11Depth(), GetPhysicalDepth(dev), vis->red_mask, vis->green_mask, vis->blue_mask);
	}
	catch(...)
	{
	    xshmimg = 0;
	    if ( xshminfo.shmaddr )
		shmdt( xshminfo.shmaddr );
	    if ( xshminfo.shmid != -1 )
		shmctl( xshminfo.shmid, IPC_RMID, 0 );
	    throw;
	}
    }
    int Create(QWidget* w, int x, int y, int xpos=0, int sub=0)
    {
        dev=(QPaintDevice*)w;

	m_w=x;
        m_h=y;
	pic_w=x;
	pic_h=y;
        width=x;
	height=y;
	this->xpos=xpos;
	Alloc();
    }
    int Free()
    {
	cerr<<"Free()"<<endl;
        if ( xshmimg == 0 )
	    return 0;
	Display *dpy = dev->x11Display();
	XSync(dpy, false);
        if ( xshmpm ) {
	    XFreePixmap( dpy, xshmpm );
	    xshmpm = 0;
        }
        XShmDetach( dpy, &xshminfo ); xshmimg->data = 0;
        qSafeXDestroyImage( xshmimg ); xshmimg = 0;
	shmdt( xshminfo.shmaddr );
        shmctl( xshminfo.shmid, IPC_RMID, 0 );
	XFreeGC(dpy, gc);

	xshminfo.shmaddr=0;
	xshminfo.shmid=0;
    }
    int Resize(int& new_w, int& new_h)
    {
	if(new_w<0)return -1;
	if(new_h<0)return -1;
	new_w&=(~7);
	new_h&=(~7);
	int xratio=(new_w+m_w-1)/m_w;
	int yratio=(new_h+m_h-1)/m_h;
	int old_xratio=(pic_w+m_w-1)/m_w;
	int old_yratio=(pic_h+m_h-1)/m_h;
	cerr<<"New size: "<<new_w<<" "<<new_h<<endl;
	pic_w=new_w;
	pic_h=new_h;
	if((xratio!=old_xratio)||(yratio!=old_yratio))
	{
	    pthread_mutex_lock(&mutex);
	    Free();
	    Alloc();
	    pthread_mutex_unlock(&mutex);
	}
    }
    virtual int Draw(QPainter* pm, const CImage* data)
    {
        char* outpic=xshmimg->data;
	if(outpic==0)
	    return 0;
	if(data==0)
	    return 0;

        Display *dpy = dev->x11Display();
    	Visual *vis = (Visual*)dev->x11Visual();

	int bit_depth=GetPhysicalDepth(dev);

	if(pthread_mutex_trylock(&mutex)==EBUSY)
	    return -1;
//	pthread_mutex_lock(&mutex);

	sync();
	const char* src;

	if((pic_h==m_h)	&& (pic_w==m_w))
	    memcpy(outpic, data->data(), data->bytes());
	else
	    zoom((short*)outpic, (short*)data->data(), pic_w, pic_h, m_w, m_h, bit_depth, m_w*((pic_w+m_w-1)/m_w));

	XShmPutImage(dpy, dev->handle(), gc,
		xshmimg, 0, 0, xpos, 0, pic_w, pic_h, true);
	
	pthread_mutex_unlock(&mutex);
	
	return 0;
    }
    virtual int sync()
    {
	Display *dpy = dev->x11Display();
        XSync(dpy, false);
	return 0;
    }
    virtual ~ShmRenderer()
    {
	Free();
	pthread_mutex_destroy(&mutex);
    }
    virtual int toggle_fullscreen(bool=false);
    virtual int drawSubtitles(const char*);

};
int ShmRenderer::toggle_fullscreen(bool maximize=false)
{
    return 0;
}
extern MyPainter* painter;
int ShmRenderer::drawSubtitles(const char* sub)
{
    return 0;
}


VideoRenderer* CreateRenderer()
{
    return new ShmRenderer;
}    
#ifdef USE_SDL
#undef __MODULE__
#define __MODULE__ "Fullscreen renderer"
int valid_filter=0;
extern PlayerWidget* g_pw;
extern QPoint menu_point;
Window wnd;
Display* dpy;
Uint32 Filter_ThreadID;

class ResizeEvent                 
{
    pthread_mutex_t mutex;
    int _w;
    int _h;
    int state;
public:
    ResizeEvent():state(0){pthread_mutex_init(&mutex, 0);}
    ~ResizeEvent(){pthread_mutex_destroy(&mutex);}
    
    void Set(int w, int h)
    {
	pthread_mutex_lock(&mutex);
	_w=w; _h=h; state=1;
	pthread_mutex_unlock(&mutex);
    }
    int Check(int& w, int& h)
    {
	pthread_mutex_lock(&mutex);
	if(!state){pthread_mutex_unlock(&mutex);return -1;}
	w=_w; h=_h; state=0;
	pthread_mutex_unlock(&mutex);
	return 0;
    }
};
static ResizeEvent resize_event;
int in_resize;
/**
 *
 * This function processes events received by SDL window.
 *
 */
char event_names[][30]=
{
	 "SDL_NOEVENT",			
       "SDL_ACTIVEEVENT",			
       "SDL_KEYDOWN",			
       "SDL_KEYUP",			
       "SDL_MOUSEMOTION",			
       "SDL_MOUSEBUTTONDOWN",		
       "SDL_MOUSEBUTTONUP",		
       "SDL_JOYAXISMOTION",		
       "SDL_JOYBALLMOTION",		
       "SDL_JOYHATMOTION",		
       "SDL_JOYBUTTONDOWN",		
       "SDL_JOYBUTTONUP",			
       "SDL_QUIT",			
       "SDL_SYSWMEVENT",			
       "SDL_EVENT_RESERVEDA",		
       "SDL_EVENT_RESERVEDB",		
       "SDL_VIDEORESIZE",			
       "SDL_EVENT_RESERVED1",		
       "SDL_EVENT_RESERVED2",		
       "SDL_EVENT_RESERVED3",		
       "SDL_EVENT_RESERVED4",		
       "SDL_EVENT_RESERVED5",		
       "SDL_EVENT_RESERVED6",		
       "SDL_EVENT_RESERVED7"
};
int EventFilter(const SDL_Event* ev)
{
    if(!valid_filter)return 1;
    int w,h;
    
    if(resize_event.Check(w,h)==0)
	if(g_pw)
	{
	    in_resize=1;
	    g_pw->PW_Resize(w,h);
	    in_resize=0;	
	}
    if((unsigned char)(ev->type) == SDL_KEYUP)
    {
	if(ev->key.keysym.sym==SDLK_ESCAPE)
	    if(g_pw)
		g_pw->PW_fullscreen();
	if(ev->key.keysym.sym==SDLK_CARET)
	    if(ev->key.keysym.mod&(KMOD_LALT|KMOD_RALT))
		if(g_pw)
		    g_pw->PW_fullscreen();
	if(g_pw)
	{
	    if(ev->key.keysym.sym==SDLK_x)
		g_pw->PW_stop_func();
	    if(ev->key.keysym.sym==SDLK_c)
		g_pw->PW_pause_func();
	    if(ev->key.keysym.sym==SDLK_v)
		g_pw->PW_play_func();
	    if(ev->key.keysym.sym==SDLK_q)
		g_pw->PW_quit_func();
	}    
    }
    if((unsigned char)(ev->type) == SDL_VIDEORESIZE)
    {
	w=ev->resize.w;
	h=ev->resize.h;
	if(g_pw)
	{
	    in_resize=1;
	    g_pw->PW_Resize(w,h);
	    in_resize=0;	
	}
    }
    if((unsigned char)(ev->type) == SDL_MOUSEBUTTONUP)
    {
	if((int)ev->button.button==3)
	    if(g_pw)
		g_pw->PW_menu_slot();
    }	
    if((unsigned char)(ev->type) == SDL_QUIT)
	if(g_pw)
	    g_pw->PW_quit_func();
    return 1;
}

class FullscreenRenderer: public VideoRenderer
{
protected:
    SDL_Surface *screen;
    int dim_x, dim_y;
    GC xgc;
    QFontMetrics fm;
    int fs;
    int m_sub;
    int zoom_divider;
    int dirty;
    Window wnd;
    pthread_mutex_t mutex;
public:
    FullscreenRenderer():screen(0), fm(RegAccess::ReadFont())
    {
	pthread_mutex_init(&mutex, 0);
    }
    int Create(QWidget* w, int width, int height, int xpos=0, int subtitles=0)
    {
	fs=SDL_RESIZABLE;
	char SDL_windowhack[256];
	dev=w;
	Display *dpy = dev->x11Display();
	::dpy=dpy;
	unsigned long mask;
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD) <0)
	    throw FATAL("Failed to init SDL");
	m_w=width;
	m_h=height;
	pic_w=width;
	pic_h=height;
	m_sub=subtitles;
	dirty=0;
	dim_x=pic_w;
        dim_y=pic_h+(subtitles?fm.height()*4:0);
	screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
	if(!screen)
	    throw FATAL("Failed to set up video mode");	    
	try
	{
	    SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);
	    SDL_SetEventFilter(EventFilter);
	    valid_filter=1;
	    SDL_WM_SetCaption("Movie window", 0);
	    
	    SDL_SysWMinfo info;

	    SDL_VERSION(&info.version);
	    SDL_GetWMInfo(&info);
	    Display *dpy=dev->x11Display();
	    XSetWindowAttributes attr;
	    XWindowAttributes src;
	    XGetWindowAttributes(dpy, info.info.x11.wmwindow, &src);
	    attr.event_mask=VisibilityChangeMask;
	    XChangeWindowAttributes(dpy, info.info.x11.wmwindow, CWEventMask, &attr);
	    XChangeWindowAttributes(dpy, info.info.x11.window, CWEventMask, &attr);
	    ::wnd=info.info.x11.wmwindow;
	    
            XGCValues gcv;

            gcv.graphics_exposures = False;
            xgc=XCreateGC(dpy, ::wnd, GCGraphicsExposures, &gcv);
            QFont font=RegAccess::ReadFont();
            XSetFont(dpy, xgc, font.handle());
	}    
	catch(...)
	{
	    SDL_QuitSubSystem(SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD);
	    throw;
	}
        return 0;
    }
    int Resize(int& new_w, int& new_h)
    {
	if(new_w<0)return -1;
	if(new_h<0)return -1;
	new_w&=(~7);
	new_h&=(~7);
	if(new_h<m_h/4)
	    return -1;
	if(new_w<m_w/4)
	    return -1;

//	cerr<<"New dimensions: "<<new_w<<" "<<new_h<<endl;
	
	if(!in_resize)
	{
	    resize_event.Set(new_w, new_h);
	    return 0;
	}	
	
    	pthread_mutex_lock(&mutex);
	SDL_LockSurface(screen);

	pic_w=new_w;
	pic_h=new_h;
	dim_x=pic_w;
        dim_y=pic_h+(m_sub?fm.height()*4:0);
	if(screen)SDL_FreeSurface(screen);
	screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);

	if(!(fs&SDL_FULLSCREEN))
	{
	    SDL_SysWMinfo info;
    	    SDL_VERSION(&info.version);
    	    SDL_GetWMInfo(&info);
	    XResizeWindow(dev->x11Display(), info.info.x11.wmwindow, dim_x, dim_y);
	}	
	SDL_UnlockSurface(screen);
	pthread_mutex_unlock(&mutex);
    }    
    virtual int Draw(QPainter* pm, const CImage* data)
    {
	if(!screen)return -1;
	if(pthread_mutex_trylock(&mutex)==EBUSY)
	    return -1;
	sync();
	SDL_LockSurface(screen);
	int bit_depth=GetPhysicalDepth(dev);
	int bpp=(bit_depth+7)/8;
	if((pic_h==m_h)	&& (pic_w==m_w))
	    memcpy(screen->pixels, data->data(), data->bytes());
	else
	    zoom((short*)screen->pixels, (short*)data->data(), pic_w, pic_h, m_w, m_h, bit_depth);
	SDL_UnlockSurface(screen);
	dirty=1;

	pthread_mutex_unlock(&mutex);
	return 0;
    }
    virtual int sync()
    {
	if(dirty)
	{
            SDL_UpdateRect(screen, 0, 0, dim_x, dim_y-(m_sub?fm.height()*4:0));
	    dirty=0;
	}    
	return 0;
    }
    virtual ~FullscreenRenderer()
    {
        XFreeGC(dpy, xgc);
    	pthread_mutex_lock(&mutex);
	SDL_SetEventFilter(0);
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	valid_filter=0;
	pthread_mutex_unlock(&mutex);
	pthread_mutex_destroy(&mutex);
    }
    virtual int toggle_fullscreen(bool maximize=false)
    {
	if(!screen)return -1;
    	pthread_mutex_lock(&mutex);
	SDL_WM_ToggleFullScreen(screen);
	fs=screen->flags;
	if(fs&SDL_FULLSCREEN)
	{
		SDL_ShowCursor(0);
	}
	else
	{
		SDL_ShowCursor(1);
	}
#ifdef HAVE_LIBXXF86VM			
	if(maximize && (fs&SDL_FULLSCREEN))
	{
	    XF86VidModeModeLine mode;
    	    int unused;
	    int w, h;
	    double ratio_x, ratio_y;
	    if ( XF86VidModeGetModeLine(dev->x11Display(), dev->x11Screen(), &unused, &mode) ) 
	    {
        	w = mode.hdisplay;
                h = mode.vdisplay-(m_sub?fm.height()*4:0);
		ratio_x=double(w)/pic_w;
		ratio_y=double(h)/pic_h;
		double ratio=(ratio_x<ratio_y)?ratio_x:ratio_y;
		w=int(ratio*pic_w);
		h=int(ratio*pic_h);
//		cerr<<mode.hdisplay<<" "<<mode.vdisplay<<" "<<w<<" "<<h<<endl;
		pthread_mutex_unlock(&mutex);
		Resize(w, h);
		return ((fs&SDL_FULLSCREEN)==0)?0:1;
    	    }
        }
#endif	
    	pthread_mutex_unlock(&mutex);
	return ((fs&SDL_FULLSCREEN)==0)?0:1;
    }
    virtual int drawSubtitles(const char* text)
    {
	static char* old_text=0;	
	if(!m_sub)return 0;
	if(!text)return 0;
	if((!old_text) && (!text[0]))return 0;

	SDL_SysWMinfo info;
	XGCValues gcv;

	SDL_VERSION(&info.version);
	SDL_GetWMInfo(&info);
	Window win;
	if(fs & SDL_FULLSCREEN)
	    win=info.info.x11.fswindow;
	else
	    win=info.info.x11.window;
	Display *dpy=dev->x11Display();
	gcv.graphics_exposures = False;
	dpy=info.info.x11.display;
	win=info.info.x11.window;
	
	if(old_text && (!strcmp(text, old_text)))
	    return 0;
	
        XSetForeground(dpy, xgc, 0x0);
	if(old_text && strcmp(text, old_text))
            XFillRectangle(dpy, win, xgc, 0, dim_y-fm.height()*4, dim_x, fm.height()*4);
        XSetForeground(dpy, xgc, 0xFFFFFFFF);
	if(old_text)delete old_text;
	old_text=new char[strlen(text)+1];
	strcpy(old_text, text);
	char* sub=new char[strlen(text)+1];
	char* sub_orig=sub;
	strcpy(sub, text);
	char* run=sub;
	int line=0;
	while(*run!=0)
	{
	    run=strchr(run, '|');
	    if(run)*run=0;
//	    XDrawString16(dpy, win, gc, 
//		dim_x/2-6*strlen(sub)/2, dim_y-65+20*line,
//		    (XChar2b*)(short*)QString::fromLocal8Bit(sub).unicode(), strlen(sub));
	    XDrawString(dpy, win, xgc, 
               (dim_x - fm.width(sub))/2, dim_y + (line-1.5)*fm.height(),
		    sub, strlen(sub));
	    if(run==0)break;
	    line++;
	    run++;
	    sub=run;
	}
	delete sub_orig;
	return 0; 
    }
};





class EnvKeeper
{
    char* str;
    static const char* sdl_var;
public:
    EnvKeeper();
    ~EnvKeeper();
};
const char* EnvKeeper::sdl_var="SDL_VIDEO_X11_NODIRECTCOLOR";

EnvKeeper::EnvKeeper()
{
    char* tmp=getenv(sdl_var);
    if(!tmp)
	str=0;
    else
    {
	str=new char[strlen(tmp)+1];
	strcpy(str, tmp);
    }	
    putenv("SDL_VIDEO_X11_NODIRECTCOLOR=1");
}

EnvKeeper::~EnvKeeper()
{
    if(!str){unsetenv(sdl_var);return;}
    putenv(str);
    delete str;
}

static EnvKeeper keeper;


class YUVRenderer: public FullscreenRenderer
{
    SDL_Overlay* _ov;
    int _fmt;
public:
    YUVRenderer(): _ov(0){}
    int Create(QWidget* w, int width, int height, int yuv_fmt, int subtitles)
    {
	FullscreenRenderer::Create(w, width, height, yuv_fmt, subtitles);
	cerr<<"Creating YUV overlay for width "<<width
	<<", height "<<height<<hex<<", fourcc 0x"<<yuv_fmt<<dec<<endl;
        _ov=SDL_CreateYUVOverlay(width, height, yuv_fmt, screen);
//	cerr<<hex<<"YUV overlay: "<<_ov<<endl;
//#if ((SDL_MAJOR_VERSION<=1) && (SDL_MINOR_VERSION<=1) && (SDL_PATCHLEVEL<=6))
//        if((!_ov) || ((yuv_fmt==fccYUY2)&&(!_ov->hw_overlay)))
	//software YUY2 converters are broken
//#else
//        if(!_ov)
//#endif
	if((!_ov)
#define _SDL_VER SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL)

#if _SDL_VER>1104
    // SDL <=1.1.5 does not have hw_overlay flag
    // since sw YUY2 was disabled, it is safe not to check it
	 || (!_ov->hw_overlay)
#endif
	)
	
	{
	    if(!_ov)
	        cerr<<"Failed to create overlay"<<endl;
	    else
		cerr<<"No hardware acceleration detected!"<<endl;
	    if(_ov)SDL_FreeYUVOverlay(_ov);
	    _ov=0;
	    return 1;
	}
	_fmt=yuv_fmt;
	return 0;
    }    
    ~YUVRenderer()
    {
	if(_ov)SDL_FreeYUVOverlay(_ov);
    }
    int Resize(int& new_w, int& new_h)
    {
	if(!_ov)return FullscreenRenderer::Resize(new_w, new_h);
	
	if(new_h<m_h/4)
	    return -1;
	if(new_w<m_w/4)
	    return -1;
//	cerr<<"New dimensions: "<<new_w<<" "<<new_h<<endl;

	if(!in_resize)
	{
	    resize_event.Set(new_w, new_h);
	    return 0;
	}

    	pthread_mutex_lock(&mutex);
	SDL_LockSurface(screen);

	pic_w=new_w;
	pic_h=new_h;
	dim_x=pic_w;
        dim_y=pic_h+(m_sub?fm.height()*4:0);
	if(screen)SDL_FreeSurface(screen);
	screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
    
	if(!(fs&SDL_FULLSCREEN))
	{
	    SDL_SysWMinfo info;
    	    SDL_VERSION(&info.version);
    	    SDL_GetWMInfo(&info);
	    XResizeWindow(dev->x11Display(), info.info.x11.wmwindow, dim_x, dim_y);
	}
	SDL_FreeYUVOverlay(_ov);
	_ov=SDL_CreateYUVOverlay(m_w, m_h, _fmt, screen);
	if(!_ov)
	{
	    cerr<<"ERROR: YUV overlay creation failed"<<endl;
	}
	SDL_UnlockSurface(screen);
	pthread_mutex_unlock(&mutex);
    }    
    int Draw(QPainter* pm, const CImage* data)
    {
	if(!screen)return -1;
	if(!data)return -1;
	if(!_ov)return FullscreenRenderer::Draw(pm, data);
	if(pthread_mutex_trylock(&mutex)==EBUSY)
	    return -1;
//	sync();
	SDL_LockYUVOverlay(_ov);
	switch(_fmt)
	{
#if _SDL_VER <= 1104
	case fccYUY2:
	case fccYV12:
	case fccIYUV:
	case fccUYVY:
	case fccYVYU:
	    memcpy(_ov->pixels, data->data(), 2*data->pixels());
	    break;
#else
	case fccYV12:
	case fccIYUV:
	    memcpy(_ov->pixels[0], data->data(), data->pixels());
	    memcpy(_ov->pixels[1], data->data()
		    +data->pixels(), data->pixels()/4);
	    memcpy(_ov->pixels[2], data->data()
		    +data->pixels()+data->pixels()/4, data->pixels()/4);
	    break;
	case fccYUY2:
	case fccUYVY:
	case fccYVYU:
	    memcpy(_ov->pixels[0], data->data(), 2*data->pixels());
	    break;
#endif	    
	}    
	SDL_UnlockYUVOverlay(_ov);
	SDL_Rect rect;
	rect.x=rect.y=0;
	rect.w=pic_w;
	rect.h=pic_h;
	SDL_DisplayYUVOverlay(_ov, &rect);
	dirty=1;
	pthread_mutex_unlock(&mutex);
	return 0;
    }
    virtual int sync()
    {
	if(!_ov)return FullscreenRenderer::sync();
	return 0;
    }
    virtual int toggle_fullscreen(bool maximize=false)
    {
	if(!_ov)return FullscreenRenderer::toggle_fullscreen(maximize);
	if(!screen)return -1;
#ifdef HAVE_LIBXXF86VM			
	if(maximize)
	{
	    XF86VidModeModeLine mode;
    	    int unused;
	    int w, h;
	    double ratio_x, ratio_y;
	    if ( XF86VidModeGetModeLine(dev->x11Display(), dev->x11Screen(), &unused, &mode) ) 
	    {
        	w = mode.hdisplay;
	        h = mode.vdisplay-(m_sub?fm.height()*4:0);
		ratio_x=double(w)/pic_w;
		ratio_y=double(h)/pic_h;
		double ratio=(ratio_x<ratio_y)?ratio_x:ratio_y;
		w=int(ratio*pic_w);
		h=int(ratio*pic_h);
//		cerr<<mode.hdisplay<<" "<<mode.vdisplay<<" "<<w<<" "<<h<<endl;
		pthread_mutex_unlock(&mutex);
		Resize(w, h);
		return ((fs&SDL_FULLSCREEN)==0)?0:1;
    	    }
        }
#endif	
	SDL_WM_ToggleFullScreen(screen);
	fs=screen->flags;
	if(fs&SDL_FULLSCREEN)
	    SDL_ShowCursor(0);
	else
	    SDL_ShowCursor(1);
	return ((fs&SDL_FULLSCREEN)==0)?0:1;
    }

};
VideoRenderer* CreateFullscreenRenderer()
{
    return new FullscreenRenderer;
}    
VideoRenderer* CreateYUVRenderer()
{
    return new YUVRenderer;
}
#endif
/*
#ifdef USE_OPENGL
struct texts {
    int w;
    int h;
    int u_w;
    int u_h;
    GLuint *texture;
    int quad_1;
    int quad_2;
    int quad_3;
    int quad_4;
    } ;

class GLFullscreenRenderer: public VideoRenderer
{

    SDL_Surface *screen;
    int dim_x, dim_y;
    int fs;
    int m_sub;
    int zoom_divider;
    int dirty;
    Window wnd;
    pthread_mutex_t mutex;
    struct texts *textures;
    GLuint *textureids;
    GLubyte* pixels;
    int textcount_w;
    int textcount_h;
    int max_texture_size;
public:
    GLFullscreenRenderer():screen(0){}


    virtual int Create(QWidget* w, int width, int height, int xpos=0, int subtitles=0)
    {



    fs=SDL_RESIZABLE | SDL_OPENGL;

    char SDL_windowhack[256];
       dev=w;
       Display *dpy = dev->x11Display();
       ::dpy=dpy;
       unsigned long mask;

  int glxMajorVersion, glxMinorVersion;


  if(SDL_Init(SDL_INIT_VIDEO ) <0)
           throw FATAL("Failed to init SDL");
    m_w=width;

         m_h=height;
    width=1024;
    height=768;
    pic_w=width;
       pic_h=height;
       m_sub=subtitles;
       dirty=0;
       dim_x=pic_w;
       dim_y=pic_h+(subtitles?80:0);
       pthread_mutex_init(&mutex, 0);
//  screen = SDL_SetVideoMode(1024, 768, 0, SDL_OPENGL);    | SDL_INIT_EVENTTHREAD
        screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
       if(screen )
       {
//         SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);
//         SDL_SetEventFilter(EventFilter);
           valid_filter=1;
           SDL_WM_SetCaption("Movie window", 0);

           SDL_SysWMinfo info;

           SDL_VERSION(&info.version);
           SDL_GetWMInfo(&info);
           Display *dpy=dev->x11Display();

      GLint maxsize=0;
      GLboolean b= false;
        SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
      glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_texture_size);
      cerr<<"Texturemaxsize="<<max_texture_size<<endl;
      glGetIntegerv(GL_RED_BITS, &maxsize);
      cerr<<"RED BITS="<<maxsize<<endl;
      glGetIntegerv(GL_BLUE_BITS, &maxsize);
      cerr<<"BLUE BITS="<<maxsize<<endl;
      glGetIntegerv(GL_GREEN_BITS, &maxsize);
      cerr<<"GREEN BITS="<<maxsize<<endl;
      glGetIntegerv(GL_ALPHA_BITS, &maxsize);
      cerr<<"ALPHA BITS="<<maxsize<<endl;
      glGetIntegerv(GL_MAX_TEXTURE_STACK_DEPTH, &maxsize);
      cerr<<"MAX_TEXTURE_STACK_DEPTH="<<maxsize<<endl;
      glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxsize);
      cerr<<"MAX_TEXTURE_UNITS="<<maxsize<<endl;
      glGetIntegerv(GL_MAX_VIEWPORT_DIMS, &maxsize);
      cerr<<"MAX_VIEWPORT_DIMS="<<maxsize<<endl;
      glGetBooleanv(GL_DOUBLEBUFFER, &b);
      if (b)
        cerr<<"DOUBLBUFFER IS OK"<<endl;
      else
        cerr<<"NO DOUBELBUFFER"<<endl;
      glGetBooleanv(GL_RGBA_MODE, &b);
      if (b)
        cerr<<"RGBA IS TRUE"<<endl;
      else
        cerr<<"RGBA IS FALSE"<<endl;
      glXQueryVersion(::dpy, &glxMajorVersion, &glxMinorVersion);
        cerr<<"glX-Version "<<glxMajorVersion<<"."<<glxMinorVersion<<endl;
      if (glXIsDirect(::dpy, glXGetCurrentContext()))
        cerr<<"Congrats, you have Direct Rendering!"<<endl;
      else {
        cerr<<"Sorry, no Direct Rendering possible!"<<endl;
        screen = false;
        }

        glEnable( GL_TEXTURE_2D );
        glViewport(0,0,pic_w, pic_h);
        if (createTextures()>0) {
          cerr<<"Fatal Error : can't create Textures"<<endl;
          screen=false;
        }
        else {
          gluOrtho2D(0, m_w, m_h, 0);
           int error = glGetError( );
          if (error)
                cerr<<"Initerror :"<<error<<endl;

//          glShadeModel(GL_SMOOTH);                   // Enables Smooth Color Shading


          glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
          glEnable( GL_TEXTURE_2D );
          if (error)
                cerr<<"Initerror :"<<error<<endl;
        }

           return 0;
       }
       if(!screen)
           throw FATAL("Failed to set up video mode");
       throw FATAL("Failed to set up fullscreen mode");

    }

    GLenum createTextures()
    {

      int skip_w=0;
      int skip_h=0;
      textcount_w=0;
      textcount_h=0;
      // This look how many textures we need in width
      while(skip_w<(m_w-8)) {
        skip_w += getPowerOf2(m_w-skip_w);
        textcount_w++;
      }
      // This look how many textures we need in height
      while(skip_h<(m_h-8)) {
        skip_h += getPowerOf2(m_h-skip_h);
        textcount_h++;
      }
      // A Backround will be set, if not the texture will not shown later
      pixels = (GLubyte*) malloc( m_w * m_h * 4);
           memset( pixels, 128, m_w * m_h * 4 );
      if (!pixels) {
        return GL_OUT_OF_MEMORY;
      }

      textures = new texts[textcount_w * textcount_h];       // Here we stores all informatio n
of the textures
      textureids = new GLuint[textcount_w * textcount_h];    // The Textureids so we can
 create all textures at once
      glGenTextures( textcount_w * textcount_h, textureids); // And here their will created
      skip_h=0;                                              // Stores the Rows to skip
      int i=0;                                               // only needed for pointerarithmetic
      for(int h=0; h<textcount_h;h++) {
        skip_w=0;                                            // the pixels in row to skip
        int text_size_h = getPowerOf2(m_h-skip_h);
        for(int w=0; w<textcount_w;w++) {
            int text_size_w = getPowerOf2(m_w-skip_w);
            (textures+i)->w = text_size_w;                   // width and height of texture
            (textures+i)->h = text_size_h;
            cerr<<"Texturesize("<<i<<") :"<<text_size_w<<"; "<<text_size_h<<endl;
            (textures+i)->u_w = skip_w;                      // Pixels zu skip in row
            (textures+i)->u_h = skip_h;                      // Rows to skip
            (textures+i)->texture = (textureids+i);
            (textures+i)->quad_1 = skip_w;                   // is the vortex of the texture,
 so we can
            (textures+i)->quad_2 = skip_h;                   // draw faster (later)
            (textures+i)->quad_3 = (textures+i)->quad_1 + (textures+i)->w;
            (textures+i)->quad_4 = (textures+i)->quad_2 + (textures+i)->h;
            skip_w +=text_size_w;

            glBindTexture( GL_TEXTURE_2D, (GLuint) (textures+i)->texture);

            glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
//GL_DECAL

            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );

            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

            glTexImage2D( GL_TEXTURE_2D,  0,  GL_RGB, (textures+i)->w, (textures+i)->h, 0,
 GL_RGBA,  GL_UNSIGNED_BYTE, pixels );
            int error = glGetError( );
            if (error) return error;
            i++;
        }
        skip_h +=text_size_h;
      }

      // Simple state setup at the end.
      glClearColor( 0.0, 0.0, 0.0, 0.0 );

      return glGetError( );
    }

    int getPowerOf2(int value)
    {
        int i,e;
        for(i=1;((i<=value) && (i<=max_texture_size)) ; i*=2)
         e=i;
        return e;
    }

    int Resize(int& new_w, int& new_h)
    {
       if(new_w<0)return -1;
       if(new_h<0)return -1;
       if(new_h<m_h/4)
           return -1;
       if(new_w<m_w/4)
           return -1;
       pic_w=new_w;
       pic_h=new_h;
       dim_x=pic_w;
       dim_y=pic_h+(m_sub?fm.height()*4:0);
       pthread_mutex_lock(&mutex);

  if(screen){ SDL_FreeSurface(screen); screen=0;}
       screen=SDL_SetVideoMode(dim_x, dim_y, 0, fs);
  if (screen)
    glViewport(0,0,pic_w, pic_h);


       pthread_mutex_unlock(&mutex);
    }






    virtual int Draw(QPainter* pm, const CImage* data)
    {
           if(!screen)return -1;
           if(pthread_mutex_trylock(&mutex)==EBUSY) return -1;
        // copy the picture and swap BGR to RGB
        memcpy(pixels, data->data(), data->bytes());
        for (int i=0;i<data->bytes();i+=3)
        {
         GLubyte tmp;
          tmp = *(pixels+i);
          *(pixels+i)=* (pixels+i+2);
          *(pixels+i+2)=tmp;
        }
        // load picture in textures and draw them
        for(int i=0; i<(textcount_w*textcount_h); i++) {
           glBindTexture( GL_TEXTURE_2D, (GLuint)(textures+i)->texture);
           glPixelStorei( GL_UNPACK_SKIP_PIXELS, (textures+i)->u_w );
           glPixelStorei( GL_UNPACK_SKIP_ROWS, (textures+i)->u_h );
           glPixelStorei( GL_UNPACK_ROW_LENGTH, m_w );

           glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0,  (textures+i)->w, (textures+i)->h, GL_RGB,
 GL_UNSIGNED_BYTE, pixels );

           glBegin( GL_QUADS );
           glTexCoord2f( 0.0, 0.0 );
           glVertex2i((textures+i)->quad_1, (textures+i)->quad_2);
           glTexCoord2f( 0.0, 1.0 );
           glVertex2i( (textures+i)->quad_1, (textures+i)->quad_4);
           glTexCoord2f( 1.0, 1.0 );
           glVertex2i( (textures+i)->quad_3, (textures+i)->quad_4 );
           glTexCoord2f( 1.0, 0.0 );
           glVertex2i((textures+i)->quad_3, (textures+i)->quad_2 );
           glEnd( );
           int error = glGetError( );
           if (error)
                cerr<<"Drawerror :"<<error<<endl;
      }
      SDL_GL_SwapBuffers();
           pthread_mutex_unlock(&mutex);
      SDL_Event event;
      while(SDL_PollEvent(&event))
        EventFilter(&event);
           return 0;
    }

    virtual int sync()
    {
       if(dirty)
       {
           SDL_UpdateRect(screen, 0, 0, dim_x, dim_y-(m_sub?fm.height()*4:0));
           dirty=0;
       }
       return 0;
    }
    virtual ~GLFullscreenRenderer()
    {
       pthread_mutex_lock(&mutex);

           free(pixels);
      SDL_SetEventFilter(0);
      SDL_QuitSubSystem(SDL_INIT_VIDEO);
      valid_filter=0;
      pthread_mutex_unlock(&mutex);
      pthread_mutex_destroy(&mutex);

    }
    virtual int toggle_fullscreen(bool maximize=false)
    {
       pthread_mutex_lock(&mutex);
       fs=fs^SDL_FULLSCREEN;
       if(screen){ SDL_FreeSurface(screen); screen=0;}
       screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), SDL_OPENGL | fs);
       if(!screen)
       {
           fs=fs^SDL_FULLSCREEN;
           screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), SDL_OPENGL |fs);
       }
  else
      glViewport(0, 0, pic_w, pic_h);

       if(fs&SDL_FULLSCREEN)
       {
               SDL_ShowCursor(0);
       }
       else
       {
               SDL_ShowCursor(1);
       }

#ifdef HAVE_LIBXXF86VM
       if(maximize && (fs&SDL_FULLSCREEN))
       {
           XF86VidModeModeLine mode;
           int unused;
           int w, h;
           double ratio_x, ratio_y;
           if ( XF86VidModeGetModeLine(dev->x11Display(), dev->x11Screen(), &unused, &mode) )
           {
               w = mode.hdisplay;
               h = mode.vdisplay-(m_sub?fm.height()*4:0);
               ratio_x=double(w)/pic_w;
               ratio_y=double(h)/pic_h;
               double ratio=(ratio_x<ratio_y)?ratio_x:ratio_y;
               w=int(ratio*pic_w);
               h=int(ratio*pic_h);
               cerr<<mode.hdisplay<<" "<<mode.vdisplay<<" "<<w<<" "<<h<<endl;
               pthread_mutex_unlock(&mutex);
               Resize(w, h);

               return ((fs&SDL_FULLSCREEN)==0)?0:1;
           }
        }
#endif
       pthread_mutex_unlock(&mutex);


       return ((fs&SDL_FULLSCREEN)==0)?0:1;

    }
    virtual int drawSubtitles(const char* text)
    {
       static char* old_text=0;
       if(!m_sub)return 0;
       if(!text)return 0;
       if((!old_text) && (!text[0]))return 0;

       SDL_SysWMinfo info;
       XGCValues gcv;

       SDL_VERSION(&info.version);
       SDL_GetWMInfo(&info);
       Window win;
       if(fs & SDL_FULLSCREEN)
           win=info.info.x11.fswindow;
       else
           win=info.info.x11.window;
       Display *dpy=dev->x11Display();
       gcv.graphics_exposures = False;
       dpy=info.info.x11.display;
       win=info.info.x11.window;

       if(old_text && (!strcmp(text, old_text)))
           return 0;

       GC gc=XCreateGC(dpy, win, GCGraphicsExposures, &gcv);
       QFont font=RegAccess::ReadFont();
       XSetFont(dpy, gc, font.handle());
       XSetForeground(dpy, gc, 0x0);
       if(old_text && strcmp(text, old_text))
           XFillRectangle(dpy, win, gc, 0, dim_y-80, dim_x, 80);
       XSetForeground(dpy, gc, 0xFFFFFFFF);
       if(old_text)delete old_text;
       old_text=new char[strlen(text)+1];
       strcpy(old_text, text);
       char* sub=new char[strlen(text)+1];
       char* sub_orig=sub;
       strcpy(sub, text);
       char* run=sub;
       int line=0;
       while(*run!=0)
       {
           run=strchr(run, '|');
           if(run)*run=0;
//         XDrawString16(dpy, win, gc,
//             dim_x/2-6*strlen(sub)/2, dim_y-65+20*line,
//                 (XChar2b*)(short*)QString::fromLocal8Bit(sub).unicode(), strlen(sub));
           XDrawString(dpy, win, gc,
               dim_x/2-6*strlen(sub)/2, dim_y-65+20*line,
                   sub, strlen(sub));
           if(run==0)break;
           line++;
           run++;
           sub=run;
       }
       delete sub_orig;
       XFreeGC(dpy, gc);
       return 0;
    }
};

VideoRenderer* CreateGLFullscreenRenderer()
{
    return new GLFullscreenRenderer;
}
#endif //USE_OPENGL
*/
#ifdef HAVE_LIBXXF86DGA
#ifdef HAVE_LIBXXF86VM
/********************************************************

	DGA video renderer
	Copyright 2000 Heinrich Langos

*********************************************************/
/*
#define INT32 __my_int32
#define INT8 __my_int8
// damn qglobal.h defines INT32 and INT8 without saying anything about signs 
// thanx to roots__ from #c for helping me out with this.

//#include "mmx_dga.cpp"
#include <X11/Xproto.h>
#include <X11/Xmd.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86dgastr.h>

#define VIDMODE_SWITCHING_DELAY 100

#include <X11/extensions/xf86vmode.h>
#include <X11/extensions/xf86vmstr.h>

class DGARenderer: public VideoRenderer
  // vm_* videomode variables
  // vp_* vievport variables
{
  Display *dpy;
  int ppl;    //pixel per line (scan line, not frame line !! )
  char* base; //frame buffer base address

  int zoom_divider;


  int  vm_count;  //number of available videomodes
  XF86VidModeModeInfo **vm_modelines;  // the modes themself
  bool vm_switched; //only used to decide if we have to switch back on quit

  int vp_width,vp_height;  // the viewport size
  int vp_flip_offset;      // memory offset to the next frame in framebuffer
  int vp_current;          // currently used viewport (0..vp_frames-1)
  const static int vp_frames=2; // the number of viewports that we use.
  int bit_depth,byte_depth;  //self explaining ...
  char *conversionbuf;  // only used if we dont render directly into the framebuffer


  void set_vidmode(int nr)
  {
    printf("switching video mode to %dx%d/%d now\n",
	   vm_modelines[nr]->hdisplay,
	   vm_modelines[nr]->vdisplay,
	   vm_modelines[nr]->dotclock);

//    XF86VidModeSwitchToMode(dpy,XDefaultScreen(dpy),vm_modelines[nr]);
    usleep(VIDMODE_SWITCHING_DELAY*1000);
  }

public:
  DGARenderer()
  {}
  int toggle_fullscreen()
  {}
  virtual int Create(QWidget* w, int x, int y, int zoom=1, int xpos=0, int sub=0)
  {
    zoom_divider=zoom;
    if(zoom!=1 && zoom!=2)throw FATAL("Unsupported zoom factor");
    vp_current=0;
    dev=(QPaintDevice*)w;
    dpy = dev->x11Display();
    m_w=x;
    m_h=y;
    width=x;
    height=y;
    Visual  *vis = (Visual*)dev->x11Visual();
    bool have_dga=false;
    int minor, major, bank, ram, flags;



    int i,bestmatch, diff_x=99999, diff_y=99999;
    XF86VidModeGetAllModeLines(dpy,XDefaultScreen(dpy),
			       &vm_count,&vm_modelines);
    for (i = 0; i < vm_count; i++) {
      if ((width/zoom_divider  <= vm_modelines[i]->hdisplay) &&
	  (height/zoom_divider <= vm_modelines[i]->vdisplay)) 
	if  ( ( (vm_modelines[i]->hdisplay - width/zoom_divider)  <= diff_x) &&
	      ( (vm_modelines[i]->vdisplay - height/zoom_divider) <= diff_y) ) {
	  bestmatch=i;
	  diff_x =vm_modelines[i]->hdisplay - width/zoom_divider ;
	  diff_y =vm_modelines[i]->vdisplay - height/zoom_divider ;
	}
    }
    i=bestmatch;
    if (i != 0 && i != vm_count) {
      set_vidmode(i);
      vm_switched = 1;
      vp_width = vm_modelines[i]->hdisplay;
      vp_height = vm_modelines[i]->vdisplay;
    } else {
      vm_switched = 0;
      vp_width = vm_modelines[0]->hdisplay;
      vp_height = vm_modelines[0]->vdisplay;
    }
//    XWarpPointer(dpy, None, RootWindow(dpy, DefaultScreen(dpy)), 
//		 0, 0, 0, 0, 0,0);
    XF86VidModeSetViewPort(dpy,XDefaultScreen(dpy),0,0);
    if(XF86DGAQueryExtension(dpy, &major, &minor))
      {	
	XF86DGAQueryDirectVideo(dpy, XDefaultScreen(dpy), &flags);
	if(flags & XF86DGADirectPresent)
	  {
	    have_dga=true;
	    XF86DGAGetVideo(dpy, DefaultScreen(dpy),  &base, &ppl, &bank, &ram);
	    if(!base)
	      return -1;

	    // This is quoted from xmms-1.0.1 dga.c
  	    // The things we must go through to get a proper depth...
	    //   If I find the idiot that thought making 32bit report 24bit
	    //   was a good idea, there may be one less `programmer' in this
	    //   world...
	    {
	      XImage *img;
	      
	      img = XGetImage(dpy, RootWindow(dpy, DefaultScreen(dpy)),
			      0, 0, 1, 1, AllPlanes, ZPixmap);
	      bit_depth = img->bits_per_pixel;
	      XDestroyImage(img);
	    }
	    byte_depth = (bit_depth+7)/8;
	    // end quote :-)
	    XF86DGAGetViewPortSize(dpy, DefaultScreen(dpy), &vp_width, &vp_height);
	  }

      }
	    
    if(have_dga) {

      
      //      printf("DGA: base:%x , ppl:%d, bank:%d, ram:%d \n",base, ppl, bank, ram );
      //      flip_offset=ppl*vp_height*byte_depth;

      vp_flip_offset=vp_height*ppl*byte_depth;
      XF86DGADirectVideo(dpy, DefaultScreen(dpy), XF86DGADirectGraphics);

      //			     XF86DGADirectMouse|
      //			     XF86DGADirectKeyb);

      XF86DGASetViewPort(dpy, DefaultScreen(dpy), 0, 0);

      //     
	CARD32 zero=0;
        for (int v=0;v<vp_flip_offset*vp_frames; v+=4) 
      	  memcpy((void*)(base+v), &zero, 4);

      return 0;
    }else {
//      conversionbuf=(char *)malloc(x*y*byte_depth);
//      if (!conversionbuf) 
//	return -1;	
    }
  }

  virtual int Draw(QPainter* pm, const CImage* data)
  {
    //    printf("Drawing %d lines to base %x ppl: %d width: %d \n",height,base,ppl, width);
    bit_depth++;
    bit_depth/=8;
    int bpl=::width*bit_depth;
//    conversionbuf+=bpl*(height-1);
    sync();
//    for(int y=::height-1; y>0; y--)
//    {
//        memcpy(conversionbuf, data, bpl);
//        data+=bpl;
//        conversionbuf-=bpl;
//    }
    for(int i=0; i<height; i++)
	memcpy((void*)(base+i*ppl*byte_depth), data->at(i), data->bpl());
  
    return 0;
  };
  virtual int sync() 
  {
    Display *dpy = dev->x11Display();
    XSync(dpy, false);
    return 0;
  };
  virtual ~DGARenderer(){
    Display *dpy = dev->x11Display();
    if (vm_switched)
       set_vidmode(0);
    XF86DGASetViewPort(dpy, DefaultScreen(dpy), 0, 0);
    XF86DGADirectVideo(dpy, DefaultScreen(dpy), 0);
//    free(conversionbuf);
  }
     virtual int drawSubtitles(const char*)
     {
        return 0;
     }
};

VideoRenderer* CreateDGARenderer()
{
    return new DGARenderer;
} 
*/   
#endif	/*HAVE_LIBXXF86VM*/
#endif  /*HAVE_LIBXXF86DGA*/
