/********************************************************

	Benchmarking tool
	Copyright 2000 Eugene Kuznetsov (divx@euro.ru)

*********************************************************/

#include <config.h>

#include <stdio.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#else
#include <stdlib.h>
#endif

#include <signal.h>
#include <iostream.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <qapp.h>
#include <qpainter.h>

#include <avifile.h>
//extern int EF_PROTECT_BELOW;
#include <aviplay.h>
#include <aviutil.h>
#include <version.h>
#include "renderer.h"
#include "playerwidget.h"

#include <wine/winreg.h>
#include <registry.h>

#define __MODULE__ "benchmark"
QWidget *m;
VideoRenderer* renderer=0;
MyPainter* painter=0;
PlayerWidget* g_pw=0;
//QApplication* qApp;
int main(int argc, char **argv)
{
//    qApp=new QApplication(argc,argv);
    QApplication a(argc,argv);

//    EF_PROTECT_BELOW=1;
    if(GetAvifileVersion()!=AVIFILE_VERSION)
    {
	cout<<"This binary was compiled for Avifile ver. "<<AVIFILE_VERSION<<", but the library is ver. "<<GetAvifileVersion()<<". Aborting."<<endl;
	return 0;
    }	

    int result;
    char* fn;
    m=new QWidget;

    if(argc==1)
    {
	cout<<"Usage: benchmark <file> "
#ifdef USE_SDL
	<<"[-sdl] [-yuv]"
#endif
	<<"[-zoom]"<<endl
	    <<"  Performs a measurement of Avifile performance for given file."<<endl
	    <<endl
	    <<"  Options: "<<endl
#ifdef USE_SDL
	    <<"  -sdl   Use SDL video renderer instead of default SHM one."<<endl
	    <<"  -yuv   Use hardware acceleration if it is available. Assumes -sdl."<<endl
#endif
	    <<"  -size x y Zoom picture to dimensions x*y"<<endl
	    <<" Description of result: "<<endl
	    <<"   \'Decompression\': Time spent on reading one video frame from file and its decompression."<<endl
	    <<"   \'Drawing\'      : Time spent on X server request."<<endl
	    <<"   \'Sync\'         : Time spent by X server on drawing the frame. "<<endl;
//	    <<"                      For DGA renderer it's zero because all drawing is performed by Avifile."<<endl
	return -1;
    }
    else
	fn=argv[1];
    int use_sdl=0;
    int use_zoom=0;
    int use_yuv=0;
    int xs, ys;
    int quality;
    int qual_key;
    int status;
    if(argc>=3)
    for(int i=2; i<argc; i++)
    {
#ifdef USE_SDL
        if(strcmp(argv[i], "-sdl")==0)
	    use_sdl=1;
	if(strcmp(argv[i], "-yuv")==0)
	{
	    use_sdl=1;
	    use_yuv=1;
	}
#endif
    if(strcmp(argv[i], "-qual")==0)
    {
	if(i+1<argc)
	{
	quality=atoi(argv[i+1]);
        CodecInfo* ci=(CodecInfo *)CodecInfo::match(fccDIV3);
	if(ci)
	    IVideoDecoder::SetExtendedAttr(*ci, "Quality", quality);
/*	    
	    if(RegCreateKeyExA(HKEY_CURRENT_USER,
	    "SOFTWARE\\Microcrap\\Scrunch", 0, 0, 0, 0, 0,
		 &qual_key, &status)==0)
	    {
		RegSetValueExA(qual_key, "Current Post Process Mode", 0, 0, &quality, 4);
		quality=-1;
		RegSetValueExA(qual_key, "Force Post Process Mode", 0, 0, &quality, 4);
		RegCloseKey(qual_key);
	    } 
*/	    
	}
    }
    if(strcmp(argv[i], "-size")==0)
    {
	if(i+2<argc)
	{
	    use_zoom=1;
	    xs=atoi(argv[i+1]);
	    ys=atoi(argv[i+2]);
	    if((xs<0)||(ys<0))use_zoom=0;
	}
    }
    }
    double read_times=0, draw_times=0, sync_times=0;
    int frames=0;
    painter=new MyPainter(m);
    IAviReadFile *file=0;
    try
    {

	file=CreateIAviReadFile(fn);
        IAviReadStream *stream=file->GetStream(0, AviStream::Video);
	
        BITMAPINFOHEADER bh;
	stream->StartStreaming();
        stream->GetDecoder()->SetDestFmt(GetPhysicalDepth(m));
        stream->GetOutputFormat(&bh, sizeof bh);
	bh.biWidth=abs(bh.biWidth);
	bh.biHeight=abs(bh.biHeight);	
        printf("Movie size: %dx%d\n", bh.biWidth, bh.biHeight);
        if(stream->GetEndPos()<112)
	    throw FATAL("Too short movie");
	m->show();
	a.setMainWidget(m);

#ifdef USE_SDL
	if(use_sdl)
	{
    	    try
	    {
		if(!use_yuv)
		    throw FATAL("Not needed");
		if(!(stream->GetDecoder()->GetCapabilities() & IVideoDecoder::CAP_YUY2))
    		    throw FATAL("YUV unsupported by decoder");
	        renderer=CreateYUVRenderer();
    		if(renderer->Create(m, bh.biWidth, bh.biHeight, fccYUY2)!=1)
	        //returns 1 if YUV is unsupported by X
		    if(stream->GetDecoder()->SetDestFmt(0, fccYUY2))//shouldn't happen
	    		throw FATAL("Error setting YUV decoder output");
	    }
	    catch(FatalError& e)
    	    {
		delete renderer;
		renderer=CreateFullscreenRenderer();
    		renderer->Create(m, bh.biWidth, bh.biHeight);
	    }
	}
	else
#endif	
	{
    	    renderer=CreateRenderer();
	    renderer->Create(m,bh.biWidth,bh.biHeight);
	}	    
        m->move(100,100);
	if(use_zoom)
	    renderer->Resize(xs, ys);
	else
	{
	    xs=bh.biWidth;
	    ys=bh.biHeight;
	}
	if(!use_sdl)
    	    m->resize(xs, ys);	    
	else
	    m->resize(20,20);

	while(!stream->Eof() && (frames<32))
	{
	    stream->ReadFrame();
	    renderer->Draw(painter, stream->GetFrame());
    	    renderer->sync();
	    frames++;
	}		
        frames=0;
	while(!stream->Eof() && (frames<80))
        {
	    int t1, t2, t3, t4;
	    t1=localcount();
	    stream->ReadFrame();
	    t2=localcount();
	    renderer->Draw(painter, stream->GetFrame());
	    t3=localcount();
	    renderer->sync();
	    t4=localcount();
	    read_times+=(t2-t1)/freq;
	    draw_times+=(t3-t2)/freq;
	    sync_times+=(t4-t3)/freq;
	    frames++;
	}		
        printf("Played %d frames in %f seconds ( max frame rate %f fps )\n",
	     frames, (read_times+draw_times+sync_times)/1000., 1000.*frames/(read_times+draw_times+sync_times));
	if(frames)
	printf("Average results:\n\t"
	    " Decompression %f ms\n\t"
	    " Drawing %f ms\n\t"
	    " Sync %f ms\n",
	    read_times/frames, draw_times/frames, sync_times/frames);
//        if(use_sdl==0)
//	{
	    printf("Average video output speed: %f Mb/s\n",
	        GetPhysicalDepth(m)/8*xs*ys*frames/sync_times/1000.);
//	}	    
        painter->end();
    }
    catch(FatalError& error)
    {
	error.Print();
    }		
    delete renderer;
    delete painter;
    delete file;
    delete m;	
    return 0;
}
