/********************************************************

	Miscellaneous image processing routines
	Copyright 2000 Eugene Kuznetsov  (divx@euro.ru)

*********************************************************/
#include <config.h>

extern void zoom_16_bpp(short* dest, const short* src, int dst_w, int dst_h, int src_w, int src_h, int xdim)
{
    int x_maj = src_w/dst_w;
    int x_min = src_w%dst_w;
    int y_maj = (src_h/dst_h)*src_w;
    int y_min = src_h%dst_h;
    int x_accum = 0;
    int y_accum = 0;
    const short *src2=src;
    int dest_delta = (xdim!=0 ? xdim-dst_w : 0);
    
       
//    cout << "(" << dst_w << "," << dst_h << "," << src_w << "," << src_h << ")" << endl;
    for(int i=0; i<dst_h; i++)
    {
       
    	for(int j=0; j<dst_w; j++)
        {
           *dest=*src;
           dest++;

           src+=x_maj;
           x_accum+=x_min;
           if(x_accum>=dst_w) {
              x_accum-=dst_w;
              src++;
           }
        }

	    dest+=dest_delta;
        src2+=y_maj;
        y_accum+=y_min;

        if(y_accum>=dst_h)
        {
            y_accum-=dst_h;
            src2+=src_w;
        }
        
        src=src2;
    }
}

extern void zoom_24_bpp(int* dest, const int* src, int dst_w, int dst_h, int src_w, int src_h, int xdim)
{
    int x_maj = src_w/dst_w;
    int x_min = src_w%dst_w;
    int y_maj = (src_h/dst_h)*src_w;
    int y_min = src_h%dst_h;
    int x_accum = 0;
    int y_accum = 0;
    const int *src2=src;
    int dest_delta = (xdim!=0 ? 3*(xdim-dst_w) : 0);
    
//    cout << "(" << dst_w << "," << dst_h << "," << src_w << "," << src_h << ")" << endl;
    for(int i=0; i<dst_h; i++)
    {
       
    	for(int j=0; j<dst_w; j++)
        {
           *dest=*src;
            dest=(int*)((char*)dest+3);

            src=(int*)((char*)src+x_maj*3);
            x_accum+=x_min;
            if(x_accum>=dst_w) {
                x_accum-=dst_w;
        	src=(int*)((char*)src+3);
            }
        }

	    dest = (int*) ((char*) dest + dest_delta);

        src2=(int*)((char*)src2+y_maj*3);
        y_accum+=y_min;
        if(y_accum>=dst_h)
        {
            y_accum-=dst_h;
    	    src2=(int*)((char*)src2+3*src_w);
        }
        src=src2;
    }
}
extern void zoom_32_bpp(int* dest, const int* src, int dst_w, int dst_h, int src_w, int src_h, int xdim)
{
    int x_maj = src_w/dst_w;
    int x_min = src_w%dst_w;
    int y_maj = (src_h/dst_h)*src_w;
    int y_min = src_h%dst_h;
    int x_accum = 0;
    int y_accum = 0;
    const int *src2=src;
    int dest_delta = (xdim!=0 ? xdim-dst_w : 0);
    
//    cout << "(" << dst_w << "," << dst_h << "," << src_w << "," << src_h << ")" << endl;
    for(int i=0; i<dst_h; i++)
    {
       
    	for(int j=0; j<dst_w; j++)
        {
           *dest=*src;
            dest++;

            src+=x_maj;
            x_accum+=x_min;
            if(x_accum>=dst_w) {
                x_accum-=dst_w;
                src++;
            }
        }

        dest+=dest_delta;

        src2+=y_maj;
        y_accum+=y_min;
        if(y_accum>=dst_h)
        {
            y_accum-=dst_h;
            src2+=src_w;
        }
        src=src2;
    }
}

extern void zoom(short* dest, const short* src, int dst_w, int dst_h, int src_w, int src_h, int bpp, int xdim=0)
{
    switch(bpp)
    {
	case 15:
	case 16:
	    return zoom_16_bpp(dest,src,dst_w,dst_h,src_w,src_h,xdim);
	case 24:
	    return zoom_24_bpp((int*)dest,(const int*)src,dst_w,dst_h,src_w,src_h,xdim);
	case 32:
	    return zoom_32_bpp((int*)dest,(const int*)src,dst_w,dst_h,src_w,src_h,xdim);
    }
}    


