#include "filters.h"
#include <qdialog.h>
#include <qmessagebox.h>
#include <qslider.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include <strstream>
string GammaFilter::name()
{
    return string("Gamma adjustment");
}

string GammaFilter::fullname()
{
    char s[128];
    sprintf(s, "Adjust gamma by %.2f", _gamma);
    return s;
}

CImage* GammaFilter::process(CImage* im, int pos)
{
    im->addref();
    if(im->depth()!=24)return im;
    if(im->yuv())im->ToRGB();
    for(int i=0; i<im->height(); i++)
    for(int j=0; j<im->width(); j++)
    {
	unsigned char* c=im->at(j,i);
	int z=c[0];
	z=int(z*_gamma);
	if(z>255)z=255;
	c[0]=z;
	z=c[1];
	z=int(z*_gamma);
	if(z>255)z=255;
	c[1]=z;
	z=c[2];
	z=int(z*_gamma);
	if(z>255)z=255;
	c[2]=z;
    }
    return im;
}

void GammaFilter::about()
{
    QMessageBox::about(0, "About Gamma adjustment Filter", "This is a sample filter");
}

void GammaFilter::config()
{
    QDialog d(0, "Gamma adjustment", true);
    d.setFixedSize(200,80);
    QSlider* slider=new QSlider(&d, "");
    slider->setTracking(false);
    QLabel* label=new QLabel(&d, "");
    label->setText("Adjust gamma ( 0..5 ):");
    label->resize(70,15);
    label->move(10,15);
    slider->setOrientation(QSlider::Horizontal);
    slider->setTickmarks(QSlider::Below);
    double old_gamma=_gamma;
    slider->move(90,15);
    slider->resize(100,25);
    slider->setRange(0,50);
    slider->setValue(int(_gamma*10));
    connect(slider, SIGNAL(valueChanged(int)), this, SLOT(valueChanged(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(100,50);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(150,50);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    if(d.exec()==QDialog::Rejected)
	_gamma=old_gamma;
    kernel->redraw();	
}
void GammaFilter::valueChanged(int new_val)
{
    _gamma=new_val/10.;
    kernel->redraw();
}
string GammaFilter::save()
{
    ostrstream s;
    s<<_gamma;
    s.flush();
    return string(s.str(), s.tellp());    
}

void GammaFilter::load(string str)
{
    istrstream s(str.c_str());
    s>>_gamma;
}


string BlurFilter::name()
{
    return "Blur filter";
}
string BlurFilter::fullname()
{
    char s[128];
    sprintf(s, "Blur with radius %d", _range);
    return s;
}
CImage* BlurFilter::process(CImage* im, int pos)
{
    im->addref();
    if(!im->yuv())
    {
	im->ToYUV();
	im->Blur(_range);
	im->ToRGB();
    }
    else	
	im->Blur(_range);
    return im;
}
void BlurFilter::about()
{
    QMessageBox::about(0, "About Blur Filter", "This is a sample MMX filter");
}
void BlurFilter::config()
{
    QDialog d(0, "Blur adjustment", true);
    d.setFixedSize(200,80);
    QSlider* slider=new QSlider(&d, "");
    slider->setTracking(false);
    QLabel* label=new QLabel(&d, "");
    label->setText("Adjust radius:");
    label->resize(70,15);
    label->move(10,15);
    slider->setOrientation(QSlider::Horizontal);
    slider->setTickmarks(QSlider::Below);
    int old_range=_range;
    slider->move(90,15);
    slider->resize(100,25);
    slider->setRange(1,8);
    slider->setValue(_range);
    connect(slider, SIGNAL(valueChanged(int)), this, SLOT(valueChanged(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(100,50);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(150,50);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    if(d.exec()==QDialog::Rejected)
	_range=old_range;
    kernel->redraw();	
}
void BlurFilter::valueChanged(int new_val)
{
    _range=new_val;
    kernel->redraw();
}
string BlurFilter::save()
{
    ostrstream s;
    s<<_range;
    s.flush();
    return string(s.str(), s.tellp());    
}

void BlurFilter::load(string str)
{
    istrstream s(str.c_str());
    s>>_range;
}


string RefractFilter::name()
{
    return "Refraction compensation";
}
string RefractFilter::fullname()
{
    char s[128];
    sprintf(s, "Compensate refraction with dist %d & power %f", _dist, _power);
    return s;
}
void RefractFilter::about()
{
    QMessageBox::about(0, "About Refraction Filter", "This is a sample filter");
}
void RefractFilter::config()
{
    QDialog d(0, "Refraction compensation", true);
    d.setFixedSize(200,120);
    QSlider* slider1=new QSlider(&d, "");
    slider1->setTracking(false);
    QSlider* slider2=new QSlider(&d, "");
    slider2->setTracking(false);
    QLabel* label1=new QLabel(&d, "");
    label1->setText("Adjust distance:");
    label1->resize(70,15);
    label1->move(10,15);
    QLabel* label2=new QLabel(&d, "");
    label2->setText("Adjust power:");
    label2->resize(70,15);
    label2->move(10,50);
    slider1->setOrientation(QSlider::Horizontal);
    slider1->setTickmarks(QSlider::Below);
    slider2->setOrientation(QSlider::Horizontal);
    slider2->setTickmarks(QSlider::Below);
    int old_dist=_dist;
    double old_power=_power;
    slider1->move(90,15);
    slider1->resize(100,25);
    slider1->setRange(1,50);
    slider1->setValue(_dist);
    slider2->move(90,50);
    slider2->resize(100,25);
    slider2->setRange(1,50);
    slider2->setValue(int(_power*100));
    connect(slider1, SIGNAL(valueChanged(int)), this, SLOT(valueChanged1(int)));    
    connect(slider2, SIGNAL(valueChanged(int)), this, SLOT(valueChanged2(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(100,80);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(150,80);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    if(d.exec()==QDialog::Rejected)
    {
	_dist=old_dist;
	_power=old_power;
    }	
    kernel->redraw();	

}


void RefractFilter::valueChanged1(int new_val)
{
    _dist=new_val;
    kernel->redraw();
}
void RefractFilter::valueChanged2(int new_val)
{
    _power=double(new_val)/100.;
    kernel->redraw();
}

CImage* RefractFilter::process(CImage* im, int pos)
{
    if(im->depth()!=24)	
	return im;
    if(!im->yuv())
	im->ToYUV();
    CImage* refractor=new CImage(im,24);
    refractor->Blur(2);
    int xd=im->width();
    int yd=im->height();
    unsigned char* data=(unsigned char*)im->data();
    unsigned char* refdata=(unsigned char*)refractor->data();    
    for(unsigned int i=xd*yd*3-3; i>=3*_dist+9+9*xd; i-=3)
    {
    	unsigned char* p=data+i;
    	unsigned char* refp=refdata+i-3-3*xd;
    	int brightness;
    	if(((i-3*_dist)%(3*xd)<2.6*xd)||((i%(3*xd)>xd)))
    			//brightness=(refp[-3*dist+0]+refp[-3*dist+1]+refp[-3*dist+2])/3;		
    			//brightness=(.114*refp[-3*dist+0]+.586*refp[-3*dist+1]+.300*refp[-3*dist+2]);		
    			//brightness=((col*)(refp-3*dist))->Y();		
    			brightness=refp[-3*_dist];
    	else brightness=0;
    	brightness=int(brightness*_power);
	//int y=(*p+power*p[-3*dist])/(1+power);
    	/*if(y<0x10)y=0x10;
    	if(y>0xef)y=0xef;*/
    	/*p[0]=max(0, unsigned char((p[0]+brightness)/(1+power)));
    	p[1]=max(0, unsigned char((p[1]+brightness)/(1+power)));
    	p[2]=max(0, unsigned char((p[2]+brightness)/(1+power)));*/
    	p[0]=(unsigned char)((p[0]+brightness)/(1+_power));
    	//p[1]=unsigned char((p[1]+brightness)/(1+power));
    	//p[2]=unsigned char((p[2]+brightness)/(1+power));
    	//*p=y;
    }
//    im->ToRGB();
    im->addref();
    return im;
}
string RefractFilter::save()
{
    ostrstream s;
    s<<_dist<<" "<<_power;
    s.flush();
    return string(s.str(), s.tellp());    
}

void RefractFilter::load(string str)
{
    istrstream s(str.c_str());
    s>>_dist>>_power;
//    char z;
//    s>>z;
//    s>>_power;
}


string NoiseFilter::name()
{
    return "Noise reduction";
}
string NoiseFilter::fullname()
{
    char s[128];
    sprintf(s, "Reduce noise with qual %d & edge radius %d", _qual, _edge);
    return s;
}
void NoiseFilter::about()
{
    QMessageBox::about(0, "About Noise reduction Filter", "This is a sample filter");
}
void NoiseFilter::config()
{
    QDialog d(0, "Noise reduction", true);
    d.setFixedSize(300,160);
    QSlider* slider1=new QSlider(&d, "");
    QSlider* slider2=new QSlider(&d, "");
    QSlider* slider3=new QSlider(&d, "");
    slider1->setTracking(false);
    slider2->setTracking(false);
    slider3->setTracking(false);
    QLabel* label1=new QLabel(&d, "");
    label1->setText("Adjust qual:");
    label1->resize(70,15);
    label1->move(10,15);
    QLabel* label2=new QLabel(&d, "");
    label2->setText("Adjust radius:");
    label2->resize(70,15);
    label2->move(10,50);
    slider1->setOrientation(QSlider::Horizontal);
    slider1->setTickmarks(QSlider::Below);
    slider2->setOrientation(QSlider::Horizontal);
    slider2->setTickmarks(QSlider::Below);
    slider3->setOrientation(QSlider::Horizontal);
    slider3->setTickmarks(QSlider::Below);
    int old_qual=_qual;
    int old_edge=_edge;
    int old_zz=_zz;
    slider1->move(90,15);
    slider1->resize(200,25);
    slider1->setRange(850,1000);
    slider1->setValue(_qual);
    slider2->move(90,50);
    slider2->resize(200,25);
    slider2->setRange(1,6);
    slider2->setValue(_edge);
    slider3->move(90,85);
    slider3->resize(200,25);
    slider3->setRange(1,6);
    slider3->setValue(_zz);
    connect(slider1, SIGNAL(valueChanged(int)), this, SLOT(valueChanged1(int)));    
    connect(slider2, SIGNAL(valueChanged(int)), this, SLOT(valueChanged2(int)));    
    connect(slider3, SIGNAL(valueChanged(int)), this, SLOT(valueChanged3(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(200,120);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(250,120);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    if(d.exec()==QDialog::Rejected)
    {
	_qual=old_qual;
	_edge=old_edge;
	_zz=old_zz;
    }	
    kernel->redraw();	

}


void NoiseFilter::valueChanged1(int new_val)
{
    _qual=new_val;
    kernel->redraw();
}
void NoiseFilter::valueChanged2(int new_val)
{
    _edge=new_val;
    kernel->redraw();
}
void NoiseFilter::valueChanged3(int new_val)
{
    _zz=new_val;
    kernel->redraw();
}

CImage* NoiseFilter::process(CImage* im, int pos)
{

    if(im->depth()!=24)
    {
	cerr<<"ERROR: not 24bpp image"<<endl;
	im->addref();
	return im;
    }	
    if(!im->yuv())
	im->ToYUV();

//	Move(move);

    double m_dQual=_qual/1000.;
    

    CImage* result=new CImage(im);
//	result->Enhance(256/m_dEnhQual);//128 ms
//	result->Move(move);

    CEdgeRgn* edges=new CEdgeRgn(im, int(256/m_dQual), _edge, false, _zz);

    CImage* blurred=new CImage(result);
    blurred->Blur(1);//53 ms
	
    CImage* blurred3=new CImage(blurred);

    blurred3->Blur(2,1);

    CEdgeRgn* v2=new CEdgeRgn(edges);
	
    v2->Normalize();//17 ms
	
    v2->Blur(1);
    CEdgeRgn* v3=new CEdgeRgn(v2);
    v3->Blur(2,1);
    
/*    
    struct yuv* yptr=(struct yuv*)result->data();
    for(int i=0; i<im->width(); i++) 
    for(int j=0; j<im->height(); j++)
	{
	    yptr[i+j*im->width()].Cb=yptr[i+j*im->width()].Cr=128;
//	    if((*v2.at(i,j))<0)
//		yptr[i+j*im->width()].Y=128;
//	    else
		yptr[i+j*im->width()].Y=*(v2.at(i,j));
	}	
*/	

    const int _xd=im->width();
    const int _yd=im->height();
    const int OFF1=-1-1*_xd;
    const int OFF2=-2-2*_xd;
    const int OFF4=-4-4*_xd;

    for(int i=4; i<_xd-4; i++)
    {
	for(int j=i+_xd*4; j<i+_xd*(_yd-4); j+=_xd)
	{
	    yuv* src=(yuv*)result->data()+j;
	    yuv* v;
	    if((*v3->offset(j+OFF4))==0)
	    {
	        v=(yuv*)blurred3->data()+j+OFF4;
	        src->Y=v->Y;
	    }
	    else 
	    if((*v2->offset(j+OFF2))==0)
	    {
		v=(yuv*)blurred->data()+j+OFF2;
	        src->Y=v->Y;
	    }
//	    else 
//	    if((*edges->offset(j))>128)
//	    {
//	        v=(yuv*)result->data()+j;
//		src->Y=128;
//		src->Cr=64;
//		src->Cb=96;
//		continue;
//	    }
//    	    else
//	    {
//		src->Y=128;
//		src->Cr=96;
//		src->Cb=64;
//		continue;
//	    }
//	    src->Cr=v->Cr;
//    	    src->Cb=v->Cb;
	}
    }
    delete edges;
    delete blurred;
    delete blurred3;
    delete v2;
    delete v3;
    result->ToRGB();
//    if((result->width()<100) || (result->height()<100)
//    || (result->width()>1000) || (result->height()>1000))
//	cerr<<"OOPS"<<endl;
    return result;
}
string NoiseFilter::save()
{
    ostrstream s;
    s<<_qual<<" "<<_edge<<" "<<_zz;
    s.flush();
    return string(s.str(), s.tellp());    
}

void NoiseFilter::load(string str)
{
    istrstream s(str.c_str());
    char z;
    cerr<<str<<endl;
//    s>>_qual>>z>>_edge>>z>>_zz;
    s>>_qual>>_edge>>_zz;
    cerr<<_qual<<" "<<_edge<<" "<<_zz<<endl;

//    s>>_qual>>" ">>_edge>>" ">>_zz;
//    s.skip(1);
//    s>>_edge;
//    s.skip(1);
//    s>>_zz;
}


string MoveFilter::name()
{
    return string("Color move");
}

string MoveFilter::fullname()
{
    char s[128];
    sprintf(s, "Move colors by %d", _delta);
    return s;
}

CImage* MoveFilter::process(CImage* im, int pos)
{
    im->addref();
    if(im->depth()!=24)return im;
    if(_delta==0)
	return im;
    if(!im->yuv())im->ToYUV();
    yuv* ptr=(yuv*)im->data();
    int csize=im->width()*im->height();
    if(_delta<0)
    {
    	for(int i=0; i<csize+_delta; i++)
    	{
	    ptr[i].Cb=ptr[i-_delta].Cb;
	    ptr[i].Cr=ptr[i-_delta].Cr;
	}
    }
    else
    {
    	for(int i=csize-1; i>=_delta; i--)
    	{
	    ptr[i].Cb=ptr[i-_delta].Cb;
	    ptr[i].Cr=ptr[i-_delta].Cr;
	}
    }
    return im;
}

void MoveFilter::about()
{
    QMessageBox::about(0, "About Color move Filter", "This is a sample filter");
}

void MoveFilter::config()
{
    QDialog d(0, "Color move adjustment", true);
    d.setFixedSize(200,80);
    QSlider* slider=new QSlider(&d, "");
    slider->setTracking(false);
    QLabel* label=new QLabel(&d, "");
    label->setText("");
    label->resize(70,15);
    label->move(10,15);
    slider->setOrientation(QSlider::Horizontal);
    slider->setTickmarks(QSlider::Below);
    int old_delta=_delta;
    slider->move(90,15);
    slider->resize(100,25);
    slider->setRange(-20,20);
    slider->setValue(_delta);
    connect(slider, SIGNAL(valueChanged(int)), this, SLOT(valueChanged(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(100,50);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(150,50);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    if(d.exec()==QDialog::Rejected)
	_delta=old_delta;
    kernel->redraw();	
}
void MoveFilter::valueChanged(int new_val)
{
    _delta=new_val;
     kernel->redraw();
}
string MoveFilter::save()
{
    ostrstream s;
    s<<_delta;
    s.flush();
    return string(s.str(), s.tellp()); 
}

void MoveFilter::load(string str)
{
    istrstream s(str.c_str());
    s>>_delta;
}
/*
CImage* SwapFilter::process(CImage* im, int pos)
{
//    if(im->yuv())im->ToRGB();
    if(!im->yuv())im->ToYUV();
    int csize=im->width()*im->height();
    yuv* ptr=(yuv*)im->data();
    for(int i=0; i<csize; i++)
    {
        int t=ptr[i].r;
	ptr[i].r=ptr[i].g;
	ptr[i].g=t;
//	col z=col(ptr[i]);
//	*(col*)(ptr+i)=z;
//	yuv z(0,0,0);
//	z.Cr=_c1*ptr[i].Cr+_c2*ptr[i].Cb;
//	z.Cb=_c3*ptr[i].Cr+_c4*ptr[i].Cb;
	int z=ptr[i].Y;
	z+=ptr[i].Cr;
	z+=ptr[i].Cb;
	z/=3;
	ptr[i].Y=z;
	ptr[i].Cr=ptr[i].Cb=128;
//	ptr[i]=z;
	
    }
    
    im->addref();
    return im;
}

void SwapFilter::config()
{
    QDialog d(0, "Noise reduction", true);
    d.setFixedSize(300,200);
    QSlider* slider1=new QSlider(&d, "");
    QSlider* slider2=new QSlider(&d, "");
    QSlider* slider3=new QSlider(&d, "");
    QSlider* slider4=new QSlider(&d, "");
//    slider1->setTracking(false);
//    slider2->setTracking(false);
//    slider3->setTracking(false);
//    slider4->setTracking(false);
    QLabel* label1=new QLabel(&d, "");
    label1->setText("Adjust qual:");
    label1->resize(70,15);
    label1->move(10,15);
    QLabel* label2=new QLabel(&d, "");
    label2->setText("Adjust radius:");
    label2->resize(70,15);
    label2->move(10,50);
    slider1->setOrientation(QSlider::Horizontal);
    slider1->setTickmarks(QSlider::Below);
    slider2->setOrientation(QSlider::Horizontal);
    slider2->setTickmarks(QSlider::Below);
    slider3->setOrientation(QSlider::Horizontal);
    slider3->setTickmarks(QSlider::Below);
    slider4->setOrientation(QSlider::Horizontal);
    slider4->setTickmarks(QSlider::Below);

    slider1->move(90,15);
    slider1->resize(200,25);
    slider1->setRange(-200,200);
    slider1->setValue(int(_c1*100));
    slider2->move(90,50);
    slider2->resize(200,25);
    slider2->setRange(-200,200);
    slider2->setValue(int(_c2*100));
    slider3->move(90,85);
    slider3->resize(200,25);
    slider3->setRange(-200,200);
    slider3->setValue(int(_c3*100));
    slider4->move(90,135);
    slider4->resize(200,25);
    slider4->setRange(-200,200);
    slider4->setValue(int(_c4*100));
    connect(slider1, SIGNAL(valueChanged(int)), this, SLOT(valueChanged1(int)));    
    connect(slider2, SIGNAL(valueChanged(int)), this, SLOT(valueChanged2(int)));    
    connect(slider3, SIGNAL(valueChanged(int)), this, SLOT(valueChanged3(int)));    
    connect(slider4, SIGNAL(valueChanged(int)), this, SLOT(valueChanged4(int)));    
    QPushButton* b_ok=new QPushButton("Ok", &d);
    b_ok->resize(40,20);
    b_ok->move(200,160);
    QPushButton* b_cancel=new QPushButton("Cancel", &d);
    b_cancel->resize(40,20);
    b_cancel->move(250,160);
    connect(b_ok, SIGNAL(clicked()), &d, SLOT(accept()));
    connect(b_cancel, SIGNAL(clicked()), &d, SLOT(reject()));
    d.exec();
    kernel->redraw();	
}
void SwapFilter::valueChanged1(int new_val)
{
    _c1=double(new_val)/100.;
    kernel->redraw();
}
void SwapFilter::valueChanged2(int new_val)
{
    _c2=double(new_val)/100.;
    kernel->redraw();
}
void SwapFilter::valueChanged3(int new_val)
{
    _c3=double(new_val)/100.;
    kernel->redraw();
}
void SwapFilter::valueChanged4(int new_val)
{
    _c4=double(new_val)/100.;
    kernel->redraw();
}
*/
CImage* ScaleFilter::process(CImage* im, int pos)
{
    if(im->depth()!=24)return im;
    int w=im->width()/2;
    int h=im->height()/2;
    col* ptr=(col*)(im->data());
    
    BitmapInfo info=*(im->get_fmt());
    adjust(info);
    CImage* result=new CImage(&info);
    col* dest=(col*)(result->data());
    for(int i=0; i<h; i++)
    {
	for(int j=0; j<w; j++)
	{	
	    *dest=*ptr;
	    dest++;
	    ptr+=2;		
	}
	ptr+=2*w;
    }        
    im->release();
    return result;
}
void ScaleFilter::adjust(BITMAPINFOHEADER& bh)
{ 
    bh.biWidth/=2;
    bh.biHeight/=2;
    bh.biSizeImage/=4;
}

#include "filters.moc"
