#include "genctl.h"
#include "recwnd.h"
#include "codecconf.h"
#include "audc.h"
#include "conf.h"
#include <aviutil.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
/* 
 *  Constructs a QtRecompressorCtl which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 */
QtRecompressorCtl::QtRecompressorCtl( QWidget* parent )
    : QtRecompressor( parent, "Qt recompressor", 0 )
{
    kernel=new RecKernel;
    wnd=0;
    ictl=0;
    no_recurs=0;

    connect(_lbFilters, SIGNAL(doubleClicked( QListBoxItem * )), this, SLOT(config_filter_clicked()));
    
    connect(_lineLowFrames, SIGNAL(textChanged(const QString&)), this, SLOT(lo_fr_changed(const QString&)));
    connect(_lineHighFrames, SIGNAL(textChanged(const QString&)), this, SLOT(hi_fr_changed(const QString&)));
    connect(_lineLowSec, SIGNAL(textChanged(const QString&)), this, SLOT(lo_sec_changed(const QString&)));
    connect(_lineHighSec, SIGNAL(textChanged(const QString&)), this, SLOT(hi_sec_changed(const QString&)));
    updateFilters();
}

/*  
 *  Destroys the object and frees any allocated resources
 */
QtRecompressorCtl::~QtRecompressorCtl()
{
    delete kernel;
    if(wnd)delete wnd;
    // no need to delete child widgets, Qt does it all for us
}

/* 
 * public slot
 */
void QtRecompressorCtl::about_filter_clicked()
{
    Filter* fi=kernel->getFilter(_lbAllFilters->currentItem());
    if(!fi)return;
    fi->about();
    fi->release();
}
/* 
 * public slot
 */

void QtRecompressorCtl::updateFilters()
{
    _lbFilters->clear();
    _lbAllFilters->clear();
    int i;
    for(i=0;;i++)
    {
	Filter* fi=kernel->getFilter(i);
	if(!fi)break;
	_lbAllFilters->insertItem(fi->name().c_str());
	fi->release();
    }
    for(i=0;;i++)
    {
	Filter* fi=kernel->getFilterAt(i);
	if(!fi)break;
	_lbFilters->insertItem(fi->fullname().c_str());
	fi->release();
    }    
}
void QtRecompressorCtl::add_filter_clicked()
{
    Filter* fi=kernel->getFilter(_lbAllFilters->currentItem());
    if(!fi)return;
    int id=kernel->addFilter(fi);
    fi->release();
    updateFilters();
}
/* 
 * public slot
 */
void QtRecompressorCtl::config_filter_clicked()
{
    Filter* fi=(*kernel)[_lbFilters->currentItem()];
    if(!fi)return;
    fi->config();
    fi->release();
    updateFilters();
}
/* 
 * public slot
 */
void QtRecompressorCtl::dest_file_clicked()
{
    QString result=QFileDialog::getSaveFileName(Registry::ReadString("QtRecomp", "SaveFile", "./dest.avi").c_str());
    if(result.isNull())
	return;
    if(kernel->setDestFile(result.ascii())!=0)
	return;
    Registry::WriteString("QtRecomp", "SaveFile", result.ascii());
    _txtDst->setText(result);
}
/* 
 * public slot
 */
void QtRecompressorCtl::details_clicked()
{
    string details=kernel->aboutStream(_comboStreams->currentItem());
    QMessageBox::information(this, "About stream", details.c_str());
}
/* 
 * public slot
 */
void QtRecompressorCtl::down_filter_clicked()
{
    kernel->moveDownFilter(_lbFilters->currentItem());
    updateFilters();
}
/* 
 * public slot
 */
void QtRecompressorCtl::load_config_clicked()
{
    QString result=QFileDialog::getOpenFileName(Registry::ReadString("QtRecomp", "OpenConf", "./test.conf").c_str());
    try
    {
	kernel->loadConfig(result.ascii());
	Registry::WriteString("QtRecomp", "OpenConf", result.ascii());
	
	_txtSrc->setText(kernel->getSrcFile());
	_txtDst->setText(kernel->getDestFile());
         _txtConfig->setText(result.ascii());

        _comboStreams->clear();
	for(int i=0; i<kernel->getStreamCount(); i++)
	    _comboStreams->insertItem(kernel->getStreamName(i).c_str());
        stream_selected(0);
//    kernel->setStreamMode(
        _comboStreams->setCurrentItem(0);
	RecKernel::Mode mode;
	if(kernel->getStreamMode(0, mode)==0)
    	    stream_mode_changed(mode);
	updateFilters();
/*   switch(mode)
	    {
	    case Ignore:
		RadioButton55->setChecked(true);
		break;
	    case Copy:
		RadioButton55_2->setChecked(true);
		break;
	    case Recompress:
		RadioButton55_3->setChecked(true);
		break;
	    }
	}    */
        RadioButton5->setChecked(true);
        stream_limit_mode_changed(0);
//	stream_mode_changed(0);
        if(!wnd)
        {
	    wnd=new SourceControlWnd(0, kernel);
	    wnd->show();
    	    ictl=new QImageControl(0);
	    ictl->show();
	}	
	else
	    wnd->update_size();
	kernel->setImageControl(ictl);
    }
    catch(FatalError& e)
    {
	QMessageBox::information(this, e.GetModule(), e.GetDesc());
    }
}
/* 
 * public slot
 */
void QtRecompressorCtl::remove_filter_clicked()
{
    kernel->removeFilter(_lbFilters->currentItem());
    updateFilters();
}
/* 
 * public slot
 */
void QtRecompressorCtl::save_config_clicked()
{
    QString result=QFileDialog::getOpenFileName(Registry::ReadString("QtRecomp", "OpenConf", "./test.conf").c_str());
    try
    {
	kernel->saveConfig(result.ascii());
	Registry::WriteString("QtRecomp", "OpenConf", result.ascii());
         _txtConfig->setText(result.ascii());
    }
    catch(FatalError& e)
    {
	QMessageBox::information(this, e.GetModule(), e.GetDesc());
    }
}
/* 
 * public slot
 */
void QtRecompressorCtl::src_file_clicked()
{
    QString result=QFileDialog::getOpenFileName(Registry::ReadString("QtRecomp", "OpenFile", "./movie.avi").c_str());
    if(result.isNull())
	return;
    if(kernel->openFile(result.ascii())!=0)
	return;
    Registry::WriteString("QtRecomp", "OpenFile", result.ascii());
    _txtSrc->setText(result);
    
    _comboStreams->clear();
    for(int i=0; i<kernel->getStreamCount(); i++)
	_comboStreams->insertItem(kernel->getStreamName(i).c_str());
    stream_selected(0);
//    kernel->setStreamMode(
    _comboStreams->setCurrentItem(0);
    RadioButton55->setChecked(true);
    RadioButton5->setChecked(true);
    stream_limit_mode_changed(0);
    stream_mode_changed(0);
//    if(vs)
//    {
//	try
//	{
//	    v_str[0].stream->StartStreaming();
//	    v_str[0].stream->ReadFrame();	
	    if(!wnd)
	    {
    		wnd=new SourceControlWnd(0, kernel);
	        wnd->show();
    		ictl=new QImageControl(0);
	        ictl->show();
	    }	
	    else
	        wnd->update_size();
	kernel->setImageControl(ictl);
//	}
//	catch(FatalError& e)
//	{
//	    e.Print();
//	    kernel->setImageControl(0);
//	}
//    }		
}
/* 
 * public slot
 */
void QtRecompressorCtl::stream_format_clicked()
{
    int stream=_comboStreams->currentItem();
    VideoEncoderInfo info;
    AudioEncoderInfo ainfo;
    WAVEFORMATEX wfmtx;
    CodecConfig* config;
    AudioCodecConfig* ac;
    if(!kernel->isStream(stream))
	return;
    switch(kernel->stream_type(stream))
    {
    case AviStream::Video:
	if(kernel->getCompress(stream, info)!=0)
	    return;//shouldn't happen.
	config=new CodecConfig(this, info);
	if(config->exec()==QDialog::Accepted)
	    kernel->setCompress(stream, config->GetInfo());
	delete config;
	break;
    case AviStream::Audio:
	if(kernel->getAudioCompress(stream, ainfo)!=0)
	    return;//shouldn't happen.
	ac=new AudioCodecConfig(this, ainfo);
	if(ac->exec()==QDialog::Accepted)
	    kernel->setAudioCompress(stream, ac->GetInfo());
	delete ac;
    }
}
/* 
 * public slot
 */
void QtRecompressorCtl::stream_limit_mode_changed(int)
{
    if(RadioButton5->isChecked())
    {
	_lineLowSec->setEnabled(true);
	_lineHighSec->setEnabled(true);
	_lineLowFrames->setEnabled(false);
	_lineHighFrames->setEnabled(false);
    }
    else
    {
    	_lineLowSec->setEnabled(false);
	_lineHighSec->setEnabled(false);
	_lineLowFrames->setEnabled(true);
	_lineHighFrames->setEnabled(true);
    }
}
/* 
 * public slot
 */
void QtRecompressorCtl::stream_mode_changed(int)
{
    int stream=_comboStreams->currentItem();
    if(!kernel->isStream(stream))
	return;
    if(RadioButton55->isChecked())
    {
	_pushFormat->setEnabled(false);
	kernel->setStreamMode(stream, RecKernel::Remove);
    }	
    else
    if(RadioButton55_2->isChecked())
    {
    	_pushFormat->setEnabled(false);
	kernel->setStreamMode(stream, RecKernel::Copy);
    }	
    else
    {
    	_pushFormat->setEnabled(true);
	kernel->setStreamMode(stream, RecKernel::Recompress);
    }	
}
/* 
 * public slot
 */
void QtRecompressorCtl::stream_selected(int s)
{
    RecKernel::Mode mode;
    if(kernel->getStreamMode(s, mode)!=0)
	return;
    int start,end;	
    char str[256];
    switch(mode)
    {
	case RecKernel::Remove:
	    RadioButton55->setChecked(true);
	    break;
	case RecKernel::Copy:
	    RadioButton55_2->setChecked(true);
	    break;
	case RecKernel::Recompress:
	    RadioButton55_3->setChecked(true);
	    break;
    }
    kernel->getSelection(s, start, end);
    sprintf(str, "%d", start);
    _lineLowFrames->setText(str);
    sprintf(str, "%d", end);
    _lineHighFrames->setText(str);
    sprintf(str, "%f", start*kernel->getFrameTime(s));
    _lineLowSec->setText(str);
    sprintf(str, "%f", end*kernel->getFrameTime(s));
    _lineHighSec->setText(str);
}
/* 
 * public slot
 */
void QtRecompressorCtl::up_filter_clicked()
{
    kernel->moveUpFilter(_lbFilters->currentItem());
    updateFilters();
}

void QtRecompressorCtl::start_recompress()
{
    if(!wnd)return;
    wnd->hide();
    hide();
    RecWindow w(this, kernel);
    w.exec();
    cerr<<"Recompress finished"<<endl;
    wnd->show();
    show();
}

void QtRecompressorCtl::lo_fr_changed(const QString& s)
{
    if(no_recurs)return;
    int stream=_comboStreams->currentItem();
    if(!kernel->isStream(stream))
	return;
    char q[128];
    sprintf(q, "%f", atoi(s.ascii())*kernel->getFrameTime(stream));
    no_recurs=1;
    _lineLowSec->setText(q);
    no_recurs=0;
    kernel->setSelectionStart(stream, atoi(s.ascii()));
} 
void QtRecompressorCtl::hi_fr_changed(const QString& s)
{
    if(no_recurs)return;
    int stream=_comboStreams->currentItem();
    if(!kernel->isStream(stream))
	return;
    char q[128];
    sprintf(q, "%f", atoi(s.ascii())*kernel->getFrameTime(stream));
    no_recurs=1;
    _lineHighSec->setText(q);
    no_recurs=0;
    kernel->setSelectionEnd(stream, atoi(s.ascii()));
}
void QtRecompressorCtl::lo_sec_changed(const QString& s)
{
    if(no_recurs)return;
    int stream=_comboStreams->currentItem();
    if(!kernel->isStream(stream))
	return;
    char q[128];
    if(kernel->getFrameTime(stream)==0)return;
    int start;
    sprintf(q, "%d", start=int(atof(s.ascii())/kernel->getFrameTime(stream)));
    no_recurs=1;
    _lineLowFrames->setText(q);
    no_recurs=0;
    kernel->setSelectionStart(stream, start);
}
void QtRecompressorCtl::hi_sec_changed(const QString& s)
{
    if(no_recurs)return;
    int stream=_comboStreams->currentItem();
    if(!kernel->isStream(stream))
	return;
    char q[128];
    if(kernel->getFrameTime(stream)==0)return;
    int end;
    sprintf(q, "%d", end=int(atof(s.ascii())/kernel->getFrameTime(stream)));
    no_recurs=1;
    _lineHighFrames->setText(q);
    no_recurs=0;
    kernel->setSelectionEnd(stream, end);
}

#include "genctl.moc"
#include "genctl_p.moc"
