#include "qimagecontrol.h"
#include <algorithm>
#include <qapp.h>
#include <qimage.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
namespace
{
    const QEvent::Type Type_LeftPix=QEvent::Type(QEvent::User);
    const QEvent::Type Type_RightPix=QEvent::Type(QEvent::User+1);
};
QImageControl::QImageControl(QWidget* parent, const char* name):QWidget(parent,name),_il(0), _ir(0)
{
    _left=new QFrame(this);
    _right=new QFrame(this);
    _left->setFrameShadow( QFrame::Sunken );
    _left->setFrameShape( QFrame::Box );
    _right->setFrameShadow( QFrame::Sunken );
    _right->setFrameShape( QFrame::Box );
    _w=_h=0;
    _il=_ir=0;
    setSize(192,144);
    pthread_mutex_init(&mutex, 0);
}

QImageControl::~QImageControl()
{
    pthread_mutex_destroy(&mutex);
    delete _il;
    delete _ir;
}

void QImageControl::setSize(int width, int height)
{
    if(height<0)height=-height;
    _left->resize(width, height);
    _right->resize(width, height);
    _left->move(20,20);
    _right->move(width+40,20);
    _w=width;
    _h=height;
    resize(2*width+60,height+40);
}
void QImageControl::setSourcePicture(const CImage* src)
{
    pthread_mutex_lock(&mutex);
    delete _il;
    _il=new CImage(src, 32);
    pthread_mutex_unlock(&mutex);
    QEvent* e=new QEvent(Type_LeftPix);
    qApp->postEvent(this, e);
}    

void QImageControl::setDestPicture(const CImage* src)
{
    pthread_mutex_lock(&mutex);
    delete _ir;
    _ir=new CImage(src, 32);
    pthread_mutex_unlock(&mutex);
    QEvent* e=new QEvent(Type_RightPix);
    qApp->postEvent(this, e);
}

void QImageControl::setPicture(const CImage* im, QFrame* frame)
{    
    if(!im)return;
    pthread_mutex_lock(&mutex);
    int bitdepth=32;
    _w=im->width();
    _h=im->height();
    int bpl=_w*4;
    QImage* i=new QImage(_w,_h,32);
    int y;
    if(!im->direction())
    for(y=0;y<_h;y++)
    {
       	QRgb *line = (QRgb *)i->scanLine(y);
	const unsigned char* vpixel= im->at(y);
	memcpy(line,vpixel,bpl);
    }
    else
    for(y=0;y<_h;y++)
    {
       	QRgb *line = (QRgb *)i->scanLine(y);
	const unsigned char* vpixel= im->at(_h-1-y);
	memcpy(line,vpixel,bpl);
    }    
    QPainter p(frame);
    p.drawImage(0,0,*i);
    delete i;
    pthread_mutex_unlock(&mutex);
}

void QImageControl::paintEvent(QPaintEvent*)
{
    setPicture(_il, _left);
    setPicture(_ir, _right);
}

bool QImageControl::event(QEvent* ev)
{
    if(ev->type()==Type_LeftPix)
    	setPicture(_il, _left);
    else
    if(ev->type()==Type_RightPix)
    	setPicture(_ir, _right);
    else
	return QWidget::event(ev);
    return true;
}
