#include "filters.h"
#include <videodecoder.h>
#include <videoencoder.h>
#include "conf.h"
#include <aviutil.h>
#include <unistd.h>
/*************************
    
    Work with filters.

*************************/
RecompressFilter::RecompressFilter(const VideoEncoderInfo& in):info(in)
{
    ve=0;
    vd=0;
    comp_frame=0;
    BITMAPINFOHEADER obh;
    try
    {
	info.header.biHeight=abs(in.header.biHeight);
	ve=IVideoEncoder::Create(info);
	obh=ve->QueryOutputFormat();
        vd=IVideoDecoder::Create(obh, 24, 1);
    	ve->SetKeyFrame(1);
//	ve->Start();
//      vd->Start();
	comp_frame=new char[abs(obh.biWidth*obh.biHeight)*4];
    }	
    catch(FatalError& e)
    {
	e.PrintAll();
	delete ve;
	delete vd;
	delete comp_frame;
	throw;
    }
}
RecompressFilter::~RecompressFilter()
{
    delete ve;
    delete vd;
    delete comp_frame;
}
CImage* RecompressFilter::process(CImage* im, int pos)
{
    BITMAPINFOHEADER bm=*(im->get_fmt());
    bm.biHeight=abs(bm.biHeight);
    if((bm.biHeight!=info.header.biHeight) || 
	(bm.biWidth!=info.header.biWidth) ||
	(bm.biBitCount!=info.header.biBitCount))
    {
	try
	{    	
	    delete ve;
	    delete vd;
	    delete comp_frame;
	    info.header=bm;
	    ve=IVideoEncoder::Create(info);
	    BITMAPINFOHEADER obh=ve->QueryOutputFormat();
	    vd=IVideoDecoder::Create(obh, 24, 1);
	    ve->SetKeyFrame(1);
    	    comp_frame=new char[abs(obh.biWidth*obh.biHeight)*4];
        }	
        catch(FatalError& e)
        {
	    e.PrintAll();
    	    delete ve;
	    delete vd;
	    delete comp_frame;
	    ve=0;
	    vd=0;
	    comp_frame=0;
	    throw;
	}
    }
    im->addref();
    int is_keyframe, size;
    ve->Start();
    vd->Start();
    int result=ve->EncodeFrame(im, comp_frame, &is_keyframe, &size);
    cerr<<"Encoded to "<<size<<" bytes, error "<<result<<endl;
    result=vd->DecodeFrame(comp_frame, size, 16);
    cerr<<result<<endl;
    CImage* tmp=vd->GetFrame();
    memcpy(im->data(), tmp->data(), im->bytes());
    tmp->release();
    vd->Stop();
    ve->Stop();
    return im;
}
string RecompressFilter::save()
{
    return string("");
}
void RecompressFilter::load(string)
{
}
//creates filter registered under id
Filter* RecKernel::getFilter(int id)
{
    switch(id)
    {
    case 0:
	return new GammaFilter(this);
    case 1:
	return new BlurFilter(this);
    case 2:
	return new RefractFilter(this);
    case 3:
	return new NoiseFilter(this);
    case 4:
	return new MoveFilter(this);
    case 5:
	return new ScaleFilter(this);	
/*    case 5:
	return new SwapFilter(this);*/
    default:
	return 0;
    }

}

//returns filter at pos id in current list
Filter* RecKernel::getFilterAt(int id)
{
    if(id<0)return 0;
    if(id>=filter_list.size())
	return 0;
    Filter* fi=filter_list[id];
    fi->addref();
    return fi;
}

//synonym
Filter* RecKernel::operator[](int id)
{
    return getFilterAt(id);
}

//size of current list
int RecKernel::getFilterCount() const
{
    return filter_list.size();
}

//adds filter in the end of list
int RecKernel::addFilter(Filter* fi)
{
    filter_list.push_back(fi);
    fi->addref();
    redraw();
    return 0;
}

int RecKernel::removeFilter(int id)
{
//    if(id<0)return 0;
//    if(id>=filter_list.size())
//	return 0;
    if(0==filter_list.size())
	return 0;
    Filter* fi=filter_list[filter_list.size()-1];
    delete fi;
    filter_list.pop_back();
    redraw();
}

int RecKernel::moveUpFilter(int id)
{
    if(id==0)return 0;
    if(id>=filter_list.size())return 0;
    swap(filter_list[id], filter_list[id-1]);
    return 0;
}

int RecKernel::moveDownFilter(int id)
{
    if(id<0)return 0;
    if(id>=(filter_list.size()-1))return 0;
    swap(filter_list[id], filter_list[id+1]);
    return 0;
}
/********************************

	Work with files.

********************************/
int RecKernel::openFile(const char* fn)
{
    if(_rf)_destruct();
    try
    {
	_rf=CreateIAviReadFile(fn);
	_readfn=fn;
	int i;
	vs=as=0;
	for(i=0; i<16; i++)
	{
	    IAviReadStream* s=_rf->GetStream(i, AviStream::Video);
	    if(!s)break;
	    full_stream fs;
	    fs.stream=s;
	    fs.startpos=0;
	    fs.endpos=s->GetEndPos();
	    fs.mode=Copy;
	    v_str.push_back(fs);
	    
//	    BITMAPINFOHEADER bh;
//	    s->GetVideoFormatInfo(&bh);
//	    i.compressor=bh.biCompression;
//	    i.quality=
	    vs++;

	    VideoEncoderInfo vi;
	    BITMAPINFOHEADER bh;
	    s->GetVideoFormatInfo((char*)&bh);
	    vi.compressor=bh.biCompression;
	    const AVIStreamHeader& h=s->GetHeader();
	    if(h.dwQuality>=0)
		vi.quality=h.dwQuality;
	    else
    		vi.quality=9500;
	    vi.keyfreq=15;
	    vfmt.push_back(vi);
	}
	for(i=0; i<16; i++)
	{
	    IAviReadStream* s=_rf->GetStream(i, AviStream::Audio);
	    if(!s)break;
	    full_stream fs;
	    fs.stream=s;
	    fs.startpos=0;
	    fs.endpos=s->GetEndPos();
	    fs.mode=Copy;
	    v_str.push_back(fs);
	    as++;
	    
	    AudioEncoderInfo ai;
	    WAVEFORMATEX wf;
	    s->GetAudioFormatInfo(&wf, 0);
	    if(wf.wFormatTag==0x55)
	    {
		ai.fmt=0x55;
		ai.bitrate=wf.nAvgBytesPerSec;
	    }	
	    else
		ai.fmt=1;
	    afmt.push_back(ai);
	}
	return 0;
    }
    catch(FatalError& e)
    {
	strLastError=e.GetModule();
	strLastError+=": ";
	strLastError+=e.GetDesc();
	return -1;
    }
}
int RecKernel::setDestFile(const char* fn)
{
    IAviWriteFile* wf;
    try
    {
	wf=CreateSegmentedFile(fn);
	_fn=fn;
	delete wf;
	return 0;
    }	
    catch(FatalError& e)
    {
	e.Print();
	return -1;
    }
}

//first all video streams, then audio
string RecKernel::aboutStream(int id) const
{
    if(!isStream(id))
	return "";
    char s[1024];
    IAviReadStream* stream=v_str[id].stream;
    if(isAudioStream(id))
    {
	WAVEFORMATEX wf;
	stream->GetAudioFormatInfo(&wf, 0);
	char* enc;
        char dummy[128];
	switch(wf.wFormatTag)
	{
	case 1:
	    enc="PCM";
	    break;
	case 2:
	    enc="MS ADPCM";
	    break;
	case 0x11://IMA ADPCM
	    enc="ADPCM";
	    break;
	case 0x31://MS GSM 6.10
	    enc="MS GSM 6.10";
	    break;
        case 0x32://MSN Audio
	    enc="MSN Audio";
	    break;
	case 0x50://MPEG Layer-1,2    
	    enc="MPEG Layer-1/2";
	    break;
	case 0x55://MPEG Layer-3
	    enc="MPEG Layer-3";
	    break;
	case 0x160:
	case 0x161://DivX audio
	    enc="DivX WMA";
	    break;
	default:
	    sprintf(dummy, "ID %d ( unsupported) ", wf.wFormatTag);
	    enc=dummy;
	    break;
	}
	const AVIStreamHeader& hdr=stream->GetHeader();
	sprintf(s, "Audio stream #%d. Length: %d samples ( %.2f seconds ).\n" 
	"Encoding: %s, %d bytes/s. Sample size %d bytes",
	    (int)id-vs, (int)stream->GetEndPos(), stream->GetEndTime(),
	    enc, hdr.dwRate, hdr.dwScale);
	return string(s);
    }
    if(isVideoStream(id))
    {
	BITMAPINFOHEADER bh;
	const AVIStreamHeader& hdr=stream->GetHeader();
	stream->GetVideoFormatInfo(&bh);
	const CodecInfo* ci=CodecInfo::match(bh.biCompression);
	string encoding;
	if(ci)encoding=ci->text;
	else
	{
	    char dummy[128];
	    int fcc[2];
	    fcc[0]=bh.biCompression;
	    fcc[1]=0;
	    sprintf(dummy, "unknown %X='%s'", bh.biCompression, fcc);
	    encoding=dummy;
	}
	sprintf(s, "Video stream #%d. %dx%dx%.2f fps. \n"
	"Length: %d frames ( %.2f seconds ). Encoding: %s.",
	(int)id, bh.biWidth, bh.biHeight, 1./stream->GetFrameTime(),
	(int)stream->GetEndPos(), stream->GetEndTime(), encoding.c_str());
    	return string(s);
    }	
}	
    
string RecKernel::lastError() const
{
    return strLastError;
}

int RecKernel::isVideoStream(int id) const
{
    if((id>=0) && (id<v_str.size()))
	return (v_str[id].stream->GetType()==AviStream::Video);
    return 0;
}
int RecKernel::isAudioStream(int id) const
{
    if((id>=0) && (id<v_str.size()))
	return (v_str[id].stream->GetType()==AviStream::Audio);
    return 0;
}
    
AviStream::StreamType RecKernel::stream_type(int id) const
{
    if((id>=0) && (id<v_str.size()))
	return v_str[id].stream->GetType();
    return AviStream::Other;	
}
    
    
int RecKernel::getCompress(int id, VideoEncoderInfo& vi) const
{
    if(!isVideoStream(id))
	return -1;
    vi=vfmt[id];
    return v_str[id].stream->GetOutputFormat(&vi.header, sizeof vi.header);
}
int RecKernel::getAudioCompress(int id, AudioEncoderInfo& wf) const
{
    if(!isAudioStream(id))
	return -1;
    wf=afmt[id-vs];
    return 0;
}

int RecKernel::setCompress(int id, const VideoEncoderInfo& vi)
{
    if(!isVideoStream(id))
	return -1;
    vfmt[id]=vi;
    if(id==0)
    {
	if(_recf)delete _recf;
        VideoEncoderInfo info;
	try
	{
	    if(getCompress(0, info)!=0)
    		_recf=0;
	    else
    		_recf=new RecompressFilter(info);
	}
	catch(FatalError& e)
	{
	    e.Print();
	}	
	redraw();
    }	    
    return 0;
}
int RecKernel::setAudioCompress(int id, const AudioEncoderInfo& wf)
{
    if(!isAudioStream(id))
	return -1;
    afmt[id-vs]=wf;
    return 0;
}

//nonzero if stream
int RecKernel::isStream(int id) const
{
    if((id>=0) && (id<v_str.size()))
	return 1;
    return 0;
}

int RecKernel::setStreamMode(int id, Mode mode)
{
    if(!isStream(id))
	return -1;
    v_str[id].mode=mode;
    return 0;
}

int RecKernel::getStreamMode(int id, Mode& mode) const
{
    if(!isStream(id))
	return -1;
    mode=v_str[id].mode;
    return 0;
}

int RecKernel::getSelection(int id, int& start, int& end) const
{
    if(!isStream(id))
	return -1;
    start=v_str[id].startpos;    
    end=v_str[id].endpos;
    return 0;
}

int RecKernel::setSelection(int id, int start, int end)
{
    if(!isStream(id))
	return -1;
    v_str[id].startpos=start;
    v_str[id].endpos=end;
}

int RecKernel::setSelectionStart(int id, int start)
{
    if(!isStream(id))
	return -1;
    v_str[id].startpos=start;
}

int RecKernel::setSelectionEnd(int id, int end)
{
    if(!isStream(id))
	return -1;
    v_str[id].endpos=end;
}

double RecKernel::getFrameTime(int id) const
{
    if(!isStream(id))
	return 0;
    return v_str[id].stream->GetFrameTime();    
}
//should set size & pictures if ctl!=0 & file is opened
void RecKernel::setImageControl(IImageControl* ctl)
{
    if(vs)
    {
	try
	{
	    v_str[0].stream->StartStreaming();
	    v_str[0].stream->ReadFrame();	
	    BITMAPINFOHEADER bh;
	    v_str[0].stream->GetVideoFormatInfo(&bh);
	    _ctl=ctl;
	    _ctl->setSize(bh.biWidth, bh.biHeight);
	    
	    if(_recf)delete _recf;
    	    VideoEncoderInfo info;
	    if(getCompress(0, info)!=0)
		_recf=0;
	    else
		_recf=new RecompressFilter(info);	

	    redraw();
	}
	catch(FatalError& e)
	{
	    e.Print();
	    return;
	}
    }    
    return;
}
    
string RecKernel::getStreamName(int id) const
{
    if(!isStream(id))
	return "";
    string s;
    if(isAudioStream(id))
    {
	s="Audio";
	id-=vs;
    }	
    else
	s="Video";
    char q[128];
    sprintf(q, " #%d", id);
    s+=q;
    return s;
}
int RecKernel::getStreamCount() const
{
    return v_str.size();
}
//in frames
//negative seeks are very slow
int RecKernel::seek(int delta)
{
    if(vs==0)return 0;
    IAviReadStream* stream=v_str[0].stream;
    int cur_pos=stream->GetPos();
    int newpos=cur_pos+delta;
    if(newpos>=stream->GetEndPos())
	return cur_pos;
    if(newpos<0)
	return cur_pos;
    int next_kern=stream->GetPrevKeyFrame(newpos);
    if((delta<0) || (next_kern>cur_pos))
	    stream->Seek(next_kern);
    while(stream->GetPos()!=newpos)
        stream->ReadFrame();
    redraw();
    return newpos;
}
//seek to nearest keyframe, return its position
int RecKernel::seek_pos(int pos)
{
    if(vs==0)return pos;
    pos=v_str[0].stream->SeekToKeyframe(pos);
    v_str[0].stream->ReadFrame();
    redraw();
    return pos;
/*    if(v_str.size()==0)return pos;
    if(vs==0)
    {
	for(int i=0; i<v_str.size(); i++)
	    v_str[i].stream->Seek(pos);
	return pos;
    }
    else
    {
	pos=v_str[0].stream->SeekToKeyframe(pos);
	for(int i=1; i<v_str.size(); i++)
	    v_str[i].stream->Seek(pos);
	redraw();
	return pos;
    }	
*/    
}

int RecKernel::pos() const
{
    if(vs==0)return 0;
    return v_str[0].stream->GetPos();
}

int RecKernel::getVideoLength() const
{
//    if(v_str.size()==0)return 0;
    if(vs==0)return 0;
//    if(v_str[0].stream->GetType()!=AviStream::Video)
//	return 0;
    return v_str[0].stream->GetEndPos();
}

/***********************************

	Actual recompression

************************************/

int RecKernel::set_callback(IRecompressCallback* rec_cb)
{
    _cb=rec_cb;
    return 0;
}

void RecKernel::_destruct()
{
    delete _rf;
    delete _wf;
    _rf=0;
    _wf=0;
    v_str.clear();
    vfmt.clear();
    afmt.clear();
}
void RecKernel::redraw()
{
    if(!_ctl)return;
    if(vs==0)return;
//    IVideoDecoder* vd;
//    IVideoEncoder* ve;
//    BITMAPINFOHEADER bh, obh;
//    v_str[0].stream->GetOutputFormat(&bh, sizeof bh);
//    try
//    {
//    }
//    catch(...)
//    {
//    }
    BITMAPINFOHEADER bh;
    v_str[0].stream->GetVideoFormatInfo(&bh);
    CImage* pix=v_str[0].stream->GetFrame();
    if(!pix)
    {
	cerr<<"ERROR: zero frame"<<endl;
	return;
    }	
    _ctl->setSourcePicture(pix);
    CImage* im=new CImage(pix);
    CImage* new_im;
    for(int i=0; i<getFilterCount(); i++)
    {
	Filter* fi=getFilterAt(i);
	new_im=fi->process(im, v_str[0].stream->GetPos());
	im->release();
	im=new_im;
    }
    
    new_im=_recf->process(im, 0);
    im->release();
    im=new_im;
    if(im->yuv())im->ToRGB();
    _ctl->setDestPicture(im);
    im->release();
}

int RecKernel::loadConfig(const char* fn)
{
    ReadConfig cfg(fn);
//    cfg.Add("SrcFile", _readfn); 
    string src=cfg.String("SrcFile");
    openFile(src.c_str());
    string dest=cfg.String("DestFile");
    setDestFile(dest.c_str());
    
//    cfg.Add("VideoStreams", vs);
//    cfg.Add("AudioStreams", as);
    int i;
    char s[128];
    for(i=0; i<v_str.size(); i++)
    {
	sprintf(s, "Stream %d start", i);
	v_str[i].startpos=cfg.Int(s);
	sprintf(s, "Stream %d end", i);
	v_str[i].endpos=cfg.Int(s);
	sprintf(s, "Stream %d mode", i);
	v_str[i].mode=(Mode)cfg.Int(s);
	if(i<vs)
	{
	    sprintf(s, "Stream %d quality", i);
	    vfmt[i].quality=cfg.Int(s);
	    sprintf(s, "Stream %d keyfreq", i);
	    vfmt[i].keyfreq=cfg.Int(s);
	    sprintf(s, "Stream %d compressor", i);
	    vfmt[i].compressor=cfg.Int(s);
//	    v_str[i].stream->GetOutputFormat(&vfmt[i].header, sizeof vfmt[i].header);
	}
	else
	{
	    sprintf(s, "Stream %d fmt", i);
	    afmt[i-vs].fmt=cfg.Int(s);
	    sprintf(s, "Stream %d bitrate", i);
	    afmt[i-vs].bitrate=cfg.Int(s);
	}
    }
    int filters=cfg.Int("Filters");
//    cfg.Add("Filters", filter_list.size());
    for(i=0; i<filters; i++)
    {
	
	sprintf(s, "Filter %d", i);
	Filter* f=getFilter(cfg.Int(s));
	if(!f)break;
	sprintf(s, "Filter %d info", i);
	string load=cfg.String(s);
	f->load(load);
	addFilter(f);
    }
    return 0;
}

int RecKernel::saveConfig(const char* fn)
{
    if(!_rf)return 0;
    WriteConfig cfg(fn);
    cfg.Add("SrcFile", _readfn); 
    cfg.Add("DestFile", _fn); 
    cfg.Add("VideoStreams", vs);
    cfg.Add("AudioStreams", as);
    int i;
    char s[128];
    for(i=0; i<v_str.size(); i++)
    {
	sprintf(s, "Stream %d start", i);
	cfg.Add(s, v_str[i].startpos);
	sprintf(s, "Stream %d end", i);
	cfg.Add(s, v_str[i].endpos);
	sprintf(s, "Stream %d mode", i);
	cfg.Add(s, (int)v_str[i].mode);
	if(i<vs)
	{
	    sprintf(s, "Stream %d quality", i);
	    cfg.Add(s, vfmt[i].quality);
	    sprintf(s, "Stream %d keyfreq", i);
	    cfg.Add(s, vfmt[i].keyfreq);
	    sprintf(s, "Stream %d compressor", i);
	    cfg.Add(s, vfmt[i].compressor);
	}
	else
	{
	    sprintf(s, "Stream %d fmt", i);
	    cfg.Add(s, afmt[i-vs].fmt);
	    sprintf(s, "Stream %d bitrate", i);
	    cfg.Add(s, afmt[i-vs].bitrate);
	}
    }
    cfg.Add("Filters", (int)filter_list.size());
    for(i=0; i<filter_list.size(); i++)
    {
	sprintf(s, "Filter %d", i);
	cfg.Add(s, filter_list[i]->id());
	sprintf(s, "Filter %d info", i);
	string save=filter_list[i]->save();
	cfg.Add(s, save);
    }
    return 0;
}
