#include "./recwnd.h"
#include <qapp.h>
#include <qprogressbar.h>
#include <qpushbutton.h>
#include <qlabel.h>


/* 
 *  Constructs a RecWindow which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
RecWindow::RecWindow( QWidget* parent,  RecKernel* kern )
    : RecWnd_p( parent, "Record window", true, 0 ), kernel(kern)
{
    progress=1;
    elapsed=estimated=1;
    fsize=0;
    kernel->set_callback(this);
    if(kernel->start_recompress()!=0)
	accept();
}

/*  
 *  Destroys the object and frees any allocated resources
 */
RecWindow::~RecWindow()
{
    // no need to delete child widgets, Qt does it all for us
}

/* 
 * public slot
 */
void RecWindow::cancel_process()
{
    kernel->stop_recompress();
    accept();
}
/* 
 * public slot
 */
void RecWindow::pause_process()
{
    if(kernel->pause_recompress()!=0)
	_pause->setText("Resume");
    else
	_pause->setText("Pause");
}
void RecWindow::Finished()
{
    qApp->postEvent(this, new QEvent(QEvent::Type(QEvent::User+1)));
//    cerr<<"Finished"<<endl;
//    exit(1);
//    accept();
}
void RecWindow::setNewState(double progress, double elapsed, double estimated, long long fsize)
{
    this->progress=progress;
    this->elapsed=elapsed;
    this->estimated=estimated;
    this->fsize=fsize;
    qApp->postEvent(this, new QEvent(QEvent::User));
}
bool RecWindow::event(QEvent* e)
{
    if(e->type()==QEvent::User)
    {
	_Progress->setProgress(int(progress*1000));
	char s[256];
        sprintf(s, "Elapsed: %d hrs %.2d min %.2d sec", int(elapsed/3600), int(elapsed/60)%60, int(elapsed)%60);
        _textElapsed->setText(s);
	sprintf(s, "Remaining: %d hrs %.2d min %.2d sec", int(estimated/3600), int(estimated/60)%60, int(estimated)%60);
        _textRemaining->setText(s);
	sprintf(s, "File size: %d Kbytes", int(fsize/1024LL));
        _textFsize->setText(s);
	fsize=(long long)(fsize/progress);
        sprintf(s, "Estimated file size: %d Kbytes", int(fsize/1024LL));
	_textEstFsize->setText(s);
        return true;
    }
    if(e->type()==QEvent::Type(QEvent::User+1))
    {
	accept();
	return true;
    }	
    return QWidget::event(e);
}
#include "recwnd.moc"
#include "recwnd_p.moc"
