#include "recompressor.h"
#include "filters.h"
#include <videodecoder.h>
#include <videoencoder.h>
#include <aviutil.h>
#include <unistd.h>
#include <qapp.h>
int RecKernel::start_recompress()
{
    rec_status=1;
    pause_status=0;
    pthread_create(&rec_thread, 0, RecKernel::RecompressThread, this);
    return 0;
}

void* RecKernel::RecompressThread(void* arg)
{
    try
    {
    
    RecKernel& kernel=*(RecKernel*)arg;

    IAviWriteFile* wf=0;
    IAviWriteStream* a_rws=0; //audio raw write stream
    IAviWriteStream* v_rws=0; //video raw write stream
    IAviAudioWriteStream* aws=0;
    IAviVideoWriteStream* vws=0;
    IAviReadStream* ars=0;
    
    int& as=kernel.as;
    int& vs=kernel.vs;
    vector<AudioEncoderInfo>& afmt=kernel.afmt;
    vector<VideoEncoderInfo>& vfmt=kernel.vfmt;

    vector<full_stream> v_str=kernel.v_str;
    string& fn=kernel._fn;
    
    IRecompressCallback* _cb=kernel._cb;
    IImageControl* _ctl=kernel._ctl;
    IAviReadFile* _rf=kernel._rf;
    
    int& rec_status=kernel.rec_status;
    int& pause_status=kernel.pause_status;


    if(as)
	if(v_str[vs].mode!=Remove)
	{
	    ars=v_str[vs].stream;
	    ars->Seek(v_str[vs].startpos);
	}	    
    IAviReadStream* vrs=0;
    if(vs)
	if(v_str[0].mode!=Remove)
        {
	    vrs=v_str[0].stream;
	    vrs->Seek(v_str[0].startpos);
	}
    if(!(vrs || ars))
	return 0;
    try
    {
	if(ars && v_str[vs].mode==Recompress)ars->StartStreaming();
	if(vrs && v_str[0].mode==Recompress)
	{
	    vrs->StopStreaming();
	    vrs->SetDirection(true);
	    vrs->StartStreaming();

	    int prev_kf=vrs->GetPrevKeyFrame(v_str[0].startpos);
	    if(prev_kf<0)prev_kf=0;
	    vrs->Seek(prev_kf);
	    while(vrs->GetPos()<v_str[0].startpos && !vrs->Eof())
		vrs->ReadFrame();
	}
    }
    catch(FatalError& e)
    {
	cerr<<"Cannot recompress this"<<endl;
	e.Print();
	return 0;
    }
    
    wf=CreateSegmentedFile(fn.c_str());
    
    WAVEFORMATEX wfmtx;
    char* ext=0;
    if(ars)
    switch(v_str[vs].mode)
    {
    case Recompress:
	ars->GetOutputFormat((char*)&wfmtx, 18);
//	if(afmt[0].fmt==0x55)
//	{
	    aws=wf->AddAudioStream(0x55, &wfmtx, 16000);
	    if(aws)
		aws->Start();
//	}    
//	else
//	    a_rws=wf->AddStream(AviStream::Audio, (char*)&wfmtx, 18, wfmtx.wFormatTag,
//		wfmtx.nAvgBytesPerSec, wfmtx.nBlockAlign);
	break;
    case Copy:
	ars->GetAudioFormatInfo((char*)&wfmtx, 0);
	if(wfmtx.cbSize>0)
	{
	    ext=new char[wfmtx.cbSize+18];
	    char* z;
	    ars->GetAudioFormatInfo((char*)&wfmtx, &ext);
	}
	if(wfmtx.nBlockAlign<=0)wfmtx.nBlockAlign=1;
	a_rws=wf->AddStream(AviStream::Audio, ext?ext:(char*)&wfmtx, ext?(wfmtx.cbSize+18):18, wfmtx.wFormatTag,
	    wfmtx.nAvgBytesPerSec, wfmtx.nBlockAlign);
	break;
    }
    delete ext;
    BITMAPINFOHEADER bh;
    if(vrs)
    switch(v_str[0].mode)
    {
    case Recompress:
	vrs->GetOutputFormat(&bh, sizeof bh);
	try
	{
    	    for(int i=0; i<kernel.getFilterCount(); i++)
	    {
	    	Filter* fi=kernel.getFilterAt(i);
		if(FormatChanger* fch=dynamic_cast<FormatChanger*>(fi))
		    fch->adjust(bh);
	    }
	    vws=wf->AddVideoStream(vfmt[0].compressor, &bh, 1000000.*vrs->GetFrameTime());
	    vws->Start();
	}
	catch(FatalError& e)
	{
	    e.Print();
	    vws=0;
	}
	break;
    case Copy:
	vrs->GetVideoFormatInfo((char*)&bh);
	v_rws=wf->AddStream(AviStream::Video, (char*)&bh, sizeof bh, 
    	    bh.biCompression, 1000000.*vrs->GetFrameTime());
	break;
    }	
    
//    	aws->Start();
//	vws->Start();
    char zz[44100];
    int comp_frame_size=abs(bh.biWidth*bh.biHeight)*4;
    char* comp_frame=new char[comp_frame_size];
    double bpf;
    if(aws)
	bpf=vrs->GetFrameTime()*(wfmtx.nSamplesPerSec*wfmtx.nChannels*wfmtx.wBitsPerSample)/8;
    else
	bpf=vrs->GetFrameTime()*wfmtx.nAvgBytesPerSec;
    long long written_audio=0;
    long long written_frames=0;
//    while((!ars->Eof()) && (!vrs->Eof()))

    long long time_start, time_current;
    time_start=longcount();
    while(1)
    {
	if((!ars)&&(!vrs))break;
	if(ars)
	{
	    if(ars->Eof())
	    {
		ars=0;
		continue;
	    }	
	    if(ars->GetPos()>=v_str[vs].endpos)
		ars=0;
	}	
	if(vrs)
	{
	    if(vrs->Eof())
	    {
		vrs=0;	
		continue;
	    }	
	    if(vrs->GetPos()>=v_str[0].endpos)
		vrs=0;
	}	
	if(!kernel.rec_status)
	    break;
	if(kernel.pause_status)
	{
	    usleep(100000);
//	    qApp->processEvents(100);
//	    qApp->wakeUpGuiThread();
	    continue;
	}
        Unsigned samples_read, bytes_read;
        if(ars && !ars->Eof())
        {
	    if(aws)
	    {
    		long long excess=(long long) (bpf*(written_frames+1)-written_audio);
//		cerr<<"Excess "<<excess<<", written audio "<<written_audio<<", written frames "<<written_frames<<", bpf "<<bpf<<endl;
		if(vrs)
		{
		    while((excess>0) && (!ars->Eof()))
	    	    {
			ars->ReadFrames(zz, (excess>sizeof zz)?excess:sizeof zz, sizeof zz, samples_read, bytes_read);
	    	        written_audio+=bytes_read;
			excess-=bytes_read;
	    	        aws->AddData(zz, bytes_read);	    	    
		    }
		}
		else
		{
			ars->ReadFrames(zz, sizeof zz, v_str[vs].endpos-ars->GetPos(), samples_read, bytes_read);
	    	        written_audio+=bytes_read;
	    	        aws->AddData(zz, bytes_read);	    	    
		}
	    }	    
	    else if(a_rws)
	    {
		long long excess=(long long) (bpf*(written_frames+1)-written_audio);
//		cerr<<"Excess "<<excess<<", written audio "<<written_audio<<", written frames "<<written_frames<<", bpf "<<bpf<<endl;
		if(vrs)
		while((excess>0) && (!ars->Eof()))
		{
		    ars->ReadDirect(zz, (excess>sizeof zz)?excess:sizeof zz, 
			((sizeof zz)/wfmtx.nBlockAlign)*wfmtx.nBlockAlign, samples_read, bytes_read);
	    	    written_audio+=bytes_read;
		    excess-=bytes_read;
		    int flags=0;
		    ars->GetFrameFlags(&flags);
	    	    a_rws->AddChunk(zz, bytes_read, flags);
		}
		else
		{
		    excess=v_str[vs].endpos-ars->GetPos();
		    if(excess>sizeof zz)excess=sizeof zz;
		    if(excess<wfmtx.nBlockAlign)
	    	    {
	    		ars=0;
	    	        continue;
		    }
		    if(excess<wfmtx.nBlockAlign)
		    {
			ars=0;
			continue;
		    }
		    ars->ReadDirect(zz, excess, 
			(excess/wfmtx.nBlockAlign)*wfmtx.nBlockAlign, samples_read, bytes_read);
	    	    written_audio+=bytes_read;
		    excess-=bytes_read;
		    int flags=0;
		    ars->GetFrameFlags(&flags);
	    	    a_rws->AddChunk(zz, bytes_read, flags);
		}
	    }
	}
	if(vrs && !vrs->Eof())
	{
	    if(vws)
	    {
		vrs->ReadFrame();
	        CImage* ptr=vrs->GetFrame();

    		if(!ptr)
		{
		    cerr<<"WARNING: zero frame"<<endl;
		    vws->AddFrame(0);
		//    vrs=0;
    		}	
		else
		{
		    CImage* im=new CImage(ptr);
		    if(written_frames%20==0)
		    {
			if(_ctl)_ctl->setSourcePicture(im);
		    }
		    	
		    for(int i=0; i<kernel.getFilterCount(); i++)
		    {
			Filter* fi=kernel.getFilterAt(i);
	    	    	CImage* new_im=fi->process(im, vrs->GetPos());
	    	    	im->release();
	    	    	im=new_im;
		    }
		    if(im->yuv())im->ToRGB();
		    if(written_frames%20==0)
		    {
			if(_ctl)
			{
			    CImage* new_im=new CImage(im);
			    if(kernel._recf)
			    {
			        CImage* new_im_2=kernel._recf->process(new_im, 0);
				new_im->release();
				new_im=new_im_2;
			    }
			    _ctl->setDestPicture(new_im);
			    new_im->release();
			}    
			time_current=longcount();
			double percent=double(written_frames)/(min((int)vrs->GetEndPos(), v_str[0].endpos)-v_str[0].startpos);
			if(_cb && percent)
			{
			cerr<<"Percent: "<<percent*100<<"%, elapsed: "<<to_float(time_current, time_start)
			    <<" seconds, remaining: "<<to_float(time_current, time_start)*(1-percent)/percent
			    <<" seconds, estimated file size:"<<2+int((wf->FileSize()-2048)/percent/1024)<<" kb"<<endl;
			    _cb->setNewState(percent, to_float(time_current, time_start),
				to_float(time_current, time_start)*(1-percent)/percent, wf->FileSize());
			}	
		    }
		    if(im->yuv())
			im->ToRGB();
		    vws->AddFrame(im);
		    im->release();
//		    qApp->processEvents(30);
//		    qApp->wakeUpGuiThread();
//		    usleep(100000);
		}    
	    }
	    else
	    {
		Unsigned samples_read, bytes_read;
		vrs->ReadDirect(comp_frame, comp_frame_size, 1, samples_read, bytes_read);
		int flags=0;
		vrs->GetFrameFlags(&flags);
		if(samples_read)
		{
		    v_rws->AddChunk(comp_frame, bytes_read, flags);
		    if(written_frames%20==0)
	    	    {
			time_current=longcount();
			double percent=double(written_frames)/(min((int)vrs->GetEndPos(), v_str[0].endpos)-v_str[0].startpos);
			if(_cb && percent)
			    _cb->setNewState(percent, to_float(time_current, time_start),
		    		to_float(time_current, time_start)*(1-percent)/percent, wf->FileSize());
		    }	    
//		    if(_cb)
//			_cb->setNewState(double(written_frames)/(min(vrs->GetEndPos(), v_str[0].endpos)-v_str[0].startpos), 0, 0);
		}
		else
		{
		    cerr<<"ERROR: Failed to read video frame"<<endl;
		    vrs=0;	
		}    
	    }	
	    written_frames++;
	}
    }
    delete wf;
//    delete ve;
//    delete vd;
    _cb->Finished();
    return 0;
    
    }
    catch(FatalError& e)
    {
	e.PrintAll();
	return 0;
    }
}
int RecKernel::pause_recompress()
{
    pause_status=!pause_status;
    return pause_status;
}
int RecKernel::stop_recompress()
{
    rec_status=0;
    pthread_join(rec_thread, 0);
    return 0;
}
