#include <qapp.h>
#include <qlabel.h>
#include <qmessagebox.h>
#include <qslider.h>
#include "dsp.h"
#include <unistd.h>
#include <avifile.h>
#include <videoencoder.h>
#include <aviutil.h>
#include "vidconf.h"
#include "capwindow.h"

#include "vidcapreg.h"

VideoRenderer* renderer=0;
int quit_flag, pause_flag, play_flag, stop_flag;


/* 
 *  Constructs a CapWindow which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 */
CapWindow::CapWindow( QWidget* parent,  const char* name, WFlags fl )
    : Form1( parent, name, fl ), m_cp(0)
{
    move(100,100);
    v4l=new v4l1if(0, RS("CapDev", "/dev/video").c_str());  
    v4l->setCapAClip(102,120,384,288);
//    v4l->addCapAClip(0,248,384,288);
    v4l->applyCapAClip(0);
    v4l->setChannel(RI("Channel", 1));
    v4l->setChannelNorm(RI("ColorMode", 0));
    
    v4l->setAudioMute(0);
    v4l->setAudioVolume(65535);//?
    v4l->setAudioBalance(100);
    v4l->setAudioMode(VIDEO_SOUND_MONO);

    connect(m_pBtn, SIGNAL(clicked()), this, SLOT(Start()));
    setFixedSize(388,388);
    connect(m_pPicProp, SIGNAL(toggled(bool)), this, SLOT(PicProperties(bool)));

    connect(m_pConfig, SIGNAL(clicked()), this, SLOT(Config()));
    connect(m_pBrightness, SIGNAL(valueChanged(int)), this, SLOT(brightChanged(int)));
    connect(m_pContrast, SIGNAL(valueChanged(int)), this, SLOT(contrastChanged(int)));
    connect(m_pColor, SIGNAL(valueChanged(int)), this, SLOT(colorChanged(int)));
    connect(m_pHue, SIGNAL(valueChanged(int)), this, SLOT(hueChanged(int)));
    
    m_pBrightness->setValue(RI("Image\\Brighness", 0));
    m_pContrast->setValue(RI("Image\\Contrast", 255));
    m_pColor->setValue(RI("Image\\Color", 255));
    m_pHue->setValue(RI("Image\\Hue", 0));

}
void CapWindow::colorChanged(int value)
{
    if(v4l)v4l->setPicColor(value);
}

void CapWindow::hueChanged(int value)
{
    if(v4l)v4l->setPicHue(value);
}

void CapWindow::PicProperties(bool show)
{
    if(show)
	setFixedSize(388,480);
	else
	setFixedSize(388,388);
}	
void CapWindow::brightChanged(int value)
{
    if(v4l)v4l->setPicBrightness(value);
}
void CapWindow::contrastChanged(int value)
{
    if(v4l)v4l->setPicConstrast(value);
}
    

/*  
 *  Destroys the object and frees any allocated resources
 */
CapWindow::~CapWindow()
{
    if(m_cp)delete m_cp;
    v4l->setCapture(0);
    v4l->setAudioMute(1);
    
    WI("Image\\Brightness", m_pBrightness->value());
    WI("Image\\Color", m_pColor->value());
    WI("Image\\Hue", m_pHue->value());
    WI("Image\\Contrast", m_pContrast->value());
    // no need to delete child widgets, Qt does it all for us
    delete v4l;
}

void CapWindow::Start()
{
/*    if(start)return;
    start=1;
    record_arg* a=new record_arg;
    a->v4l=v4l;
    a->start=&start;
    pthread_create(&thread, 0, record, a);
    connect(&m_timer, SIGNAL(timeout()), this, SLOT(Update()));
    m_timer.start(1000);*/
    
    if(m_cp)return;
    m_cp=new CaptureProcess(v4l,
	RS("FileName", "./movie.avi"),
	RI("IsSegmented", 1)?RI("SegmentSize", 1000*1000):-1,
	RI("Compressor", fccIV50),
	RI("Quality", 9500),
	RI("Keyframe", 15),
	(enum CaptureProcess::Sound_Freqs)
	    (RI("HaveAudio", 1)?RI("Frequency", 0):CaptureProcess::NoAudio),
	(enum CaptureProcess::Sample_Sizes)
	    RI("SampleSize", 0),
	(enum CaptureProcess::Sound_Chans)
	    RI("SndChannels", 0),
	(enum CaptureProcess::Resolutions)
	    RI("Resolutions", 0),
	RI("LimitTime", 0)?RI("TimeLimit", 3600):-1,
	RI("LimitSize", 0)?RI("SizLimit", 2000000):-1,
	RF("FPS", 25.)
	);
    if(RI("IsSegmented",1))
    {
	m_pSegment->setEnabled(true);
	connect(m_pSegment, SIGNAL(clicked()), this, SLOT(segment()));
    }	
    m_pBtn->setText("Stop");
    
    disconnect(m_pBtn, SIGNAL(clicked()), this, SLOT(Start()));
    connect(m_pBtn, SIGNAL(clicked()), this, SLOT(Stop()));
    connect(&m_timer, SIGNAL(timeout()), this, SLOT(Update()));
    mess.sync();
    mess.clear();
    mess.seekg(0);
    m_cp->setMessenger(mess);
    m_timer.start(1000);
    
//    record(a);
}
void CapWindow::Stop()
{
    if(!m_cp)return;
    disconnect(m_pBtn, SIGNAL(clicked()), this, SLOT(Stop()));
    disconnect(&m_timer, SIGNAL(timeout()), this, SLOT(Update()));
    connect(m_pBtn, SIGNAL(clicked()), this, SLOT(Start()));
    if(m_pSegment->isEnabled())
    {
	disconnect(m_pSegment, SIGNAL(clicked()), this, SLOT(Stop()));
	m_pSegment->setEnabled(false);
    }	
    m_pBtn->setText("Start");
    m_timer.stop();
    delete m_cp;
    m_cp=0;
    usleep(100000);
//    string inf;
//    mess.get(inf, (char)0);
    v4l->setCapture(0);
    QMessageBox::information(this, "Capture finished", mess.str());
    v4l->setCapture(1);
}
void CapWindow::Update()
{
    if(!m_cp)return;
    char str[256];
    int cap, vid, comp, quit;
    m_cp->getState(cap,vid,comp,quit);
    sprintf(str, "Captured: %d frames, dropped in capture: %d, dropped in encoder: %d", cap, vid, comp);
    m_pStatus->setText(str);
    if(quit)Stop();    
}
void CapWindow::moveEvent( QMoveEvent * e)
{
    static int state=0;
    if(state==0){state=1;return;}
    v4l->setCapture(0);
//    usleep(100000);
    v4l->setCapAClip(e->pos().x()+2, e->pos().y()+20, 384,288);
    v4l->applyCapAClip(0);
//    usleep(100000);
    v4l->setCapture(1);
//    usleep(100000);
    known_pos=1;
}
void CapWindow::showEvent( QShowEvent * )
{
    if(!known_pos)return;
    v4l->setCapture(1);
    usleep(100000);
}
void CapWindow::hideEvent( QHideEvent * )
{
    v4l->setCapture(0);
    usleep(100000);
}
void CapWindow::focusInEvent( QFocusEvent * )
{
//    if(!known_pos)return;
//    v4l->setCapture(1);
//    usleep(100000);
}
void CapWindow::focusOutEvent( QFocusEvent * )
{
//    v4l->setCapture(0);
//    usleep(100000);
}
void CapWindow::enterEvent( QEvent * )
{
}
void CapWindow::leaveEvent( QEvent * )
{
}

void CapWindow::Config()
{
    if(m_cp)return;
//    CodecConfig conf(this, "Select codec", true);
//    if(conf.exec())
//	cerr<<"New quality "<<conf.GetInfo().quality<<endl;
    v4l->setCapture(0);
    VidConfig conf(this, v4l);
    conf.exec();
    v4l->setChannel(RI("Channel", 1));
    v4l->setChannelNorm(RI("ColorMode", 0));
    v4l->setCapture(1);
}
void CapWindow::segment()
{
    if(!m_cp)return;
    m_cp->Segment();
}
#include "capwindow_p.moc"
#include "capwindow.moc"
