/********************************************************

	Video renderers
	Copyright 2000 Eugene Kuznetsov  (divx@euro.ru)

*********************************************************/
#include <config.h>

#include <qwidget.h>
#ifdef __FreeBSD__
#include <machine/param.h>
#include <sys/types.h>
#endif
#include <sys/ipc.h>
#include <sys/shm.h>                               

#include <malloc.h>
#include <pthread.h>
#include "renderer.h"
#include "RegAccess.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#ifdef HAVE_LIBXXF86DGA
#include <X11/extensions/xf86dga.h>
#endif


#ifdef USE_SDL
#include <SDL.h>
#include <SDL_video.h>
#include <SDL_syswm.h>
#endif

#include <except.h>

#define __MODULE__ "VideoRenderer"
int width, height;
//code taken from kwintv
int GetPhysicalDepth(QPaintDevice* dev)
{
        Display *dpy = dev->x11Display();
	int n, pixmap_bits;
	XPixmapFormatValues *pf=XListPixmapFormats(dpy, &n);
	for ( int i = 0; i<n; i++)
	{
		if(pf[i].depth==QColor::numBitPlanes())
		pixmap_bits=pf[i].bits_per_pixel;
	}
	return pixmap_bits;
}

inline static void qSafeXDestroyImage( XImage *x )
{
    if ( x->data ) {
	free( x->data );
	x->data = 0;
    }
    XDestroyImage( x );
}



class ShmRenderer : public VideoRenderer
{
    bool	       xshminit;
    XShmSegmentInfo xshminfo;
    XImage	      *xshmimg;
    Pixmap	       xshmpm;
    GC 			gc;
    int zoom_divider;
    int xpos;
public:
    ShmRenderer() : xshminit(0), xshmimg(0), xshmpm(0){}
    int Create(QWidget* w, int x, int y, int zoom=1, int xpos=0, int sub=0)
    {
        static int major, minor;
	static Bool pixmaps_ok;
        dev=(QPaintDevice*)w;
        Display *dpy = dev->x11Display();
        int dd	 = dev->x11Depth();
        Visual *vis	 = (Visual*)dev->x11Visual();
	zoom_divider=zoom;
	if(zoom!=1 && zoom!=2)throw FATAL("Unsupported zoom factor");
	m_w=x;
        m_h=y;
        width=x;
	height=y;
	this->xpos=xpos;

	printf("Creating SHM renderer for window %x, width %d, height %d\n", w, x, y);
	printf("Server vendor %s, release %d\n", ServerVendor(dpy), VendorRelease(dpy));
	
	try
	{
    	    XGCValues xcg;
	    xcg.graphics_exposures=false;
	    gc=XCreateGC(dpy, dev->handle(), GCGraphicsExposures, &xcg);
	    if ( !XShmQueryVersion(dpy, &major, &minor, &pixmaps_ok) )
		throw FATAL("MIT SHM extension not supported");

	    bool ok;
    	    xshminfo.shmid = shmget( IPC_PRIVATE,
    			     width*height*4/zoom/zoom,
    			     IPC_CREAT | 0777 );
	    ok = xshminfo.shmid != -1;
	    if(!ok)
		throw FATAL("Can't get shared memory segment");
    	
	    xshminfo.shmaddr = (char*)shmat( xshminfo.shmid, 0, 0 );
	    ok = xshminfo.shmaddr != 0;
	    xshminfo.readOnly = FALSE;

    	    if ( !ok )
		throw FATAL("Can't attach shared memory segment");
	    
	    ok = XShmAttach( dpy, &xshminfo );
	    if ( !ok )
		throw FATAL("XShmAttach failed");
	    
    	    xshmimg = XShmCreateImage( dpy, vis, dd, ZPixmap, xshminfo.shmaddr, &xshminfo, width/zoom, height/zoom );

	    if ( !xshmimg )
	        throw FATAL("Can't create shared image");

    	    printf("Video device: %d bpp ( physical %d ), color masks %X,%X,%X\n",
		dev->x11Depth(), GetPhysicalDepth(dev), vis->red_mask, vis->green_mask, vis->blue_mask);
    	    return 0;
	}
	catch(...)
	{
	    xshmimg = 0;
	    if ( xshminfo.shmaddr )
		shmdt( xshminfo.shmaddr );
	    if ( xshminfo.shmid != -1 )
		shmctl( xshminfo.shmid, IPC_RMID, 0 );
	    throw;
	}
    }
    virtual int Draw(QPainter* pm, const char* data)
    {
        char* outpic=xshmimg->data;
	if(outpic==0)
	    return 0;
	if(data==0)
	    return 0;

        Display *dpy = dev->x11Display();
    	Visual *vis = (Visual*)dev->x11Visual();

	int bit_depth=GetPhysicalDepth(dev);
	sync();
	const char* src;
	
	if(zoom_divider==1)
	    memcpy(outpic, data, ::height*::width*bit_depth/8);
	else
	    zoom_2((short*)outpic, (short*)data, ::width, ::height, bit_depth);
	XShmPutImage(dpy, dev->handle(), gc,
		xshmimg, 0, 0, xpos, 0, ::width/zoom_divider, ::height/zoom_divider, true);
	return 0;
    }
    virtual int sync()
    {
	Display *dpy = dev->x11Display();
        XSync(dpy, false);
	return 0;
    }
    virtual ~ShmRenderer()
    {
        if ( xshmimg == 0 )
	    return;
	Display *dpy = dev->x11Display();
	XSync(dpy, false);
        if ( xshmpm ) {
	    XFreePixmap( dpy, xshmpm );
	    xshmpm = 0;
        }
        XShmDetach( dpy, &xshminfo ); xshmimg->data = 0;
        qSafeXDestroyImage( xshmimg ); xshmimg = 0;
	shmdt( xshminfo.shmaddr );
        shmctl( xshminfo.shmid, IPC_RMID, 0 );
    }
    virtual int toggle_fullscreen();
    virtual int drawSubtitles(const char*);

};
int ShmRenderer::toggle_fullscreen()
{
    return 0;
}
extern MyPainter* painter;
int ShmRenderer::drawSubtitles(const char* sub)
{
    return 0;
}


VideoRenderer* CreateRenderer()
{
    return new ShmRenderer;
}    
#ifdef USE_SDL
#undef __MODULE__
#define __MODULE__ "Fullscreen renderer"
extern int showconf_flag, stop_flag, pause_flag, play_flag, quit_flag;
int valid_filter=0;
extern VideoRenderer* renderer;
int EventFilter(const SDL_Event* ev)
{
    if(!valid_filter)return 1;
//    cerr<<"Event type "<<(int)(unsigned char)(ev->type)<<endl;
    if((unsigned char)(ev->type) == SDL_KEYUP)
    {
//	cerr<<"Key "<<int(ev->key.keysym.sym)<<endl;
	if(ev->key.keysym.sym==SDLK_ESCAPE)
	    if(renderer)
		renderer->toggle_fullscreen();
	if(ev->key.keysym.sym==SDLK_CARET)
	    if(ev->key.keysym.mod&(KMOD_LALT|KMOD_RALT))
		if(renderer)
		    renderer->toggle_fullscreen();
	if(ev->key.keysym.sym==SDLK_x)
	    stop_flag=1;
	if(ev->key.keysym.sym==SDLK_c)
	    pause_flag=1;
	if(ev->key.keysym.sym==SDLK_v)
	    play_flag=1;
    }
    if((unsigned char)(ev->type) == SDL_QUIT)
	quit_flag=1;
    return 1;
}

class FullscreenRenderer: public VideoRenderer
{
    SDL_Surface *screen;
    int dim_x, dim_y;
    int fs;
    int m_sub;
    int zoom_divider;
    int dirty;
    Window wnd;
    pthread_mutex_t mutex;
public:
    FullscreenRenderer():screen(0){}
    virtual int Create(QWidget* w, int width, int height, int zoom=1, int xpos=0, int subtitles=0)
    {
	fs=0;
	char SDL_windowhack[256];
	dev=w;
	Display *dpy = dev->x11Display();
	unsigned long mask;
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD) <0)
	    throw FATAL("Failed to init SDL");
	m_w=width;
	m_h=height;
	zoom_divider=zoom;
	m_sub=subtitles;
	dirty=0;
	dim_x=width/zoom;
	dim_y=height/zoom+(subtitles?80:0);
	pthread_mutex_init(&mutex, 0);
	screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
	if(screen )
	{
//	    SDL_WM_GrabInput(SDL_GRAB_ON);
//	    SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);
//	    SDL_EventState(SDL_KEYUP, SDL_ENABLE);
//	    if(SDL_ShowCursor(0)!=0)
//		SDL_ShowCursor(1);
	    SDL_SetEventFilter(EventFilter);
	    valid_filter=1;
	    SDL_WM_SetCaption("Movie window", 0);
	    return 0;
	}    
	if(!screen)
	    throw FATAL("Failed to set up video mode");	    
	throw FATAL("Failed to set up fullscreen mode");
    }    
    virtual int Draw(QPainter* pm, const char* data)
    {
	if(!screen)return -1;
	pthread_mutex_lock(&mutex);
	sync();
	SDL_LockSurface(screen);
	int bit_depth=GetPhysicalDepth(dev);
	int bpp=(bit_depth+7)/8;
	if(zoom_divider==1)
	    memcpy(screen->pixels, data, m_w*m_h*bit_depth/8);
	else
	    zoom_2((short*)screen->pixels, (short*)data, m_w, m_h, bit_depth);
	SDL_UnlockSurface(screen);
	dirty=1;
	pthread_mutex_unlock(&mutex);
	return 0;
    }
    virtual int sync()
    {
	if(dirty)
	{
	    SDL_UpdateRect(screen, 0, 0, dim_x, dim_y-(m_sub?80:0));
	    dirty=0;
	}    
	return 0;
    }
    virtual ~FullscreenRenderer()
    {
    	pthread_mutex_lock(&mutex);
	SDL_SetEventFilter(0);
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	valid_filter=0;
	pthread_mutex_unlock(&mutex);
	pthread_mutex_destroy(&mutex);
    }
    virtual int toggle_fullscreen()
    {
    	pthread_mutex_lock(&mutex);

	fs=fs^SDL_FULLSCREEN;
	if(screen)SDL_FreeSurface(screen);
	screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
	if(!screen)
	{
	    fs=fs^SDL_FULLSCREEN;
	    screen=SDL_SetVideoMode(dim_x, dim_y, GetPhysicalDepth(dev), fs);
	}    
	if(fs&SDL_FULLSCREEN)
	{
		SDL_ShowCursor(0);
	}
	else
	{
		SDL_ShowCursor(1);
	}		
	pthread_mutex_unlock(&mutex);

	return (fs==0)?0:1;
    }
    virtual int drawSubtitles(const char* text)
    {
	static char* old_text=0;	
	if(!m_sub)return 0;
	if(!text)return 0;
	if((!old_text) && (!text[0]))return 0;

	SDL_SysWMinfo info;
	XGCValues gcv;

	SDL_VERSION(&info.version);
	SDL_GetWMInfo(&info);
	Window win;
	if(fs & SDL_FULLSCREEN)
	    win=info.info.x11.fswindow;
	else
	    win=info.info.x11.window;
	Display *dpy=dev->x11Display();
	gcv.graphics_exposures = False;
	dpy=info.info.x11.display;
	win=info.info.x11.window;
	
	if(old_text && (!strcmp(text, old_text)))
	    return 0;
	
	GC gc=XCreateGC(dpy, win, GCGraphicsExposures, &gcv);
	QFont font=RegAccess::ReadFont();
	XSetFont(dpy, gc, font.handle());
	XSetForeground(dpy, gc, 0x0);
	if(old_text && strcmp(text, old_text))
	    XFillRectangle(dpy, win, gc, 0, dim_y-80, dim_x, 80);
	XSetForeground(dpy, gc, 0xFFFFFFFF);
	if(old_text)delete old_text;
	old_text=new char[strlen(text)+1];
	strcpy(old_text, text);
	char* sub=new char[strlen(text)+1];
	char* sub_orig=sub;
	strcpy(sub, text);
	char* run=sub;
	int line=0;
	while(*run!=0)
	{
	    run=strchr(run, '|');
	    if(run)*run=0;
	    XDrawString16(dpy, win, gc, 
		dim_x/2-6*strlen(sub)/2, dim_y-65+20*line,
		    (XChar2b*)(short*)QString::fromLocal8Bit(sub).unicode(), strlen(sub));
	    if(run==0)break;
	    line++;
	    run++;
	    sub=run;
	}
	delete sub_orig;
	XFreeGC(dpy, gc);
	return 0; 
    }
};
VideoRenderer* CreateFullscreenRenderer()
{
    return new FullscreenRenderer;
}    
#endif
